/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.dependency;

import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.resolver.filter.ArtifactFilter;
import org.apache.maven.model.Dependency;
import org.apache.maven.project.MavenProject;

import java.util.Set;


   /**
* Interface for classes that perform dependency resolution on maven projects. Contains functions for acquiring
* the results of a dependency resolution, as well as functions for filtering what dependencies are returned
* in those results.
*/
   public interface IDependencyResolver {
   /**
    * Adds a dependency filter that will be used to conditionally exclude dependencies as they are resolved
    * @param filter the filter to add
    */
   void addDependencyFilter ( ArtifactFilter filter );

   /**
    * Resolves dependencies for the given project and returns the results of that resolution.
    * @param project the project which should have its dependencies resolved
    * @return the dependencies of the project, transitive and direct, as a {@link Set} of {@link Dependency} objects.
    */
   Set<Dependency> resolve (MavenProject project);

   /**
    * Resolves dependencies for the given project hierarchy and returns the results of that resolution.
    * @param hierarchy the hierarchy which should have its dependencies resolved
    * @return the dependencies of the hierarchy, transitive and direct, as a {@link Set} of {@link Dependency} objects.
    */
   Set<Dependency> resolve(MavenProjectHierarchy hierarchy);

   /**
    * Returns the root node of a tree of {@link DependencyNode}s representing the dependencies
    * of the entire project hierarchy along with their sources.
    * @param hierarchy the project hierarchy to gather dependencies from
    * @return the root node of a tree of {@link DependencyNode}s representing the dependencies
    * of the entire project hierarchy along with their sources
    */
   DependencyNode resolveDependencyHierarchy(MavenProjectHierarchy hierarchy);

   /**
    * Returns the root node of a tree of {@link DependencyNode}s representing the dependencies
    * of the entire project along with their sources.
    * @param project the project to gather dependencies from
    * @return the root node of a tree of {@link DependencyNode}s representing the dependencies
    * of the entire project along with their sources
    */
   DependencyNode resolveDependencyHierarchy(MavenProject project);

   /**
    * Resolves the dependencies for the given project and returns them as a set of {@link Artifact}s
    * @param project the project to resolve dependencies for
    * @return the dependencies for the given project and returns them as a set of {@link Artifact}s
    */
   public Set<Artifact> resolveDependencyArtifacts (MavenProject project);

   /**
    * Resolves the dependencies for the given project hierarchy and returns them as a set of {@link Artifact}s
    * @param hierarchy the project hierarchy to resolve dependencies for
    * @return the dependencies for the given project and returns them as a set of {@link Artifact}s
    */
   public Set<Artifact> resolveDependencyArtifacts (MavenProjectHierarchy hierarchy);
   }
