/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.dependency;

import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import com.thetus.oss.maven.util.*;
import com.thetus.oss.maven.util.DependencySorter;
import com.thetus.oss.maven.util.ArtifactSorter;
import com.thetus.oss.maven.util.DependencyHelper;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.resolver.filter.ArtifactFilter;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Dependency;
import org.apache.maven.project.DefaultProjectBuildingRequest;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuilder;
import org.apache.maven.project.ProjectBuildingException;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

/**
 * Dependency resolution class for maven that uses a process closely tied to maven's own internal process.
 * Can be used
 * to resolve dependencies as either {@link Dependency} objects or {@link Artifact} objects. Unless there
 * is some compelling cause, getting {@link Artifact} objects is encouraged since that is the native
 * representation, and contains more information as well.
 *
 * @author Samson "Zigurd" Bock - sbock@thetus.com
 */
public class MavenDependencyResolver extends ADependencyResolver implements IDependencyResolver {
    /**
     * Passthrough to parent constructor {@link ADependencyResolver#ADependencyResolver(org.apache.maven.project.ProjectBuilder, org.apache.maven.execution.MavenSession)} 
     * @param builder used to construct maven projects from pom.xml files and artifacts
     * @param session current maven session, used to acquire project building requests for project construction
     */
    public MavenDependencyResolver(@NotNull ProjectBuilder builder, @NotNull MavenSession session) {
        super(builder, session);
    }

    /**
     * Resolves dependencies for a single project and returns the result as a set of alphabetically ordered {@link Artifact}s.
     * Internally, this function makes use of the {@link ProjectBuilder} to construct a second representation of the supplied
     * project that is ensured to have dependency resolution enabled. This should allow the function to work regardless of the
     * source of the MavenProject, so long as its pom file is valid, dependencies should be accurately resolved.
     * @param project the maven project to resolve dependencies for
     * @return an alphabetically ordered set of artifacts corresponding to the transitive dependencies of the supplied project
     */
    @Override
    public Set<Artifact> resolveDependencyArtifacts (@NotNull MavenProject project) {
        Set<Artifact> artifacts = new TreeSet<Artifact>(new ArtifactSorter());

        if ( project.getArtifacts() != null ) {
            artifacts.addAll(project.getArtifacts());
        }
        DefaultProjectBuildingRequest request = new DefaultProjectBuildingRequest(m_session.getProjectBuildingRequest());
        request.setResolveDependencies(true);
        try {
            MavenProject resolvedProject = m_projectBuilder.build(project.getFile(),request).getProject();
            if ( resolvedProject != null ) {
                artifacts.addAll(resolvedProject.getArtifacts() );
            }
        } catch (ProjectBuildingException e) {
            if ( project.getArtifacts() != null ) {
                artifacts.addAll(project.getArtifacts());
            } else {
                System.out.println("[ERROR] problem encountered while trying to resolve project, backup attempt to use raw project failed");
                throw new IllegalStateException("Could not build project based on raw project, raw project did not have resolved dependencies");
            }
        }

        List<Artifact> artifactsToRemove = new ArrayList<Artifact>();
        for ( Artifact art : artifacts ) {
            for (ArtifactFilter filter : m_filters ) {
                if ( ! filter.include(art) ) {
                    artifactsToRemove.add(art);
                    break;
                }
            }
        }

        for ( Artifact art : artifactsToRemove ) {
            artifacts.remove(art);
        }

        return artifacts; 
    }

    /**
     * Resolves the dependencies for a project by means of {@link MavenDependencyResolver#resolveDependencyArtifacts(org.apache.maven.project.MavenProject)}
     * then converts the result to a set of {@link Dependency} objects.
     * @param project the project to resolve
     * @return an alphabetically ordered set of {@link Dependency} objects representing the transitive dependencies of the supplied project
     */
    @Override
    public Set<Dependency> resolve(@NotNull MavenProject project) {
        Set<Dependency> ret = new TreeSet<Dependency>(new DependencySorter());
        Set<Artifact> artifacts = resolveDependencyArtifacts(project);

        for ( Artifact art : artifacts ) {
            ret.add( DependencyHelper.convertArtifactToDependency(art) );
        }
        return ret;
    }

    /**
     * Resolves the dependencies for a project hierarchy by means of {@link MavenDependencyResolver#resolveDependencyArtifacts(org.apache.maven.project.MavenProject)}
     * @param hierarchy the project hierarchy to resolve
     * @return an alphabetically ordered set of {@link Dependency} objects representing the transitive dependencies of the supplied project hierarchy
     */
    @Override
    public Set<Artifact> resolveDependencyArtifacts (@NotNull MavenProjectHierarchy hierarchy) {
        List<MavenProject> mavenProjectList = hierarchy.getProjectList();
        Set<Artifact> artifacts = new TreeSet<Artifact>(new ArtifactSorter());
        for ( MavenProject project : mavenProjectList ) {
            artifacts.addAll(resolveDependencyArtifacts(project));
        }
        return artifacts;
    }

    /**
     * Resolves the dependencies for a project hierarchy by means of {@link MavenDependencyResolver#resolveDependencyArtifacts(MavenProjectHierarchy)}
     * then converts the results into {@link Dependency}s.
     * @param hierarchy the maven project hierarchy to resolve
     * @return an alphabetically sorted set of {@link Dependency} objects
     */
    @Override
    public Set<Dependency> resolve(MavenProjectHierarchy hierarchy) {
        Set<Dependency> ret = new TreeSet<Dependency>(new DependencySorter());
        Set<Artifact> artifacts = resolveDependencyArtifacts(hierarchy);

        for ( Artifact art : artifacts ) {
            ret.add( DependencyHelper.convertArtifactToDependency(art) );
        }
        return ret;
    }

    /**
     * Not implemented
     * @param hierarchy unused
     * @return null
     */
    @Override
    public DependencyNode resolveDependencyHierarchy(MavenProjectHierarchy hierarchy) {
        System.out.println("[WARNING] MavenDependencyResolver does not support the resolveDependencyHierarchy() functions");
        return new DependencyNode(null);
    }

    /**
     * No implemented
     * @param project unused
     * @return null
     */
    @Override
    public DependencyNode resolveDependencyHierarchy(MavenProject project) {
        System.out.println("[WARNING] MavenDependencyResolver does not support the resolveDependencyHierarchy() functions");
        return new DependencyNode(null);
    }
}
