/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.execution;

import com.thetus.oss.maven.hierarchy.DefaultMavenProjectHierarchyBuilder;
import com.thetus.oss.maven.hierarchy.IMavenProjectHierarchyBuilder;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import org.apache.maven.Maven;
import org.apache.maven.execution.MavenExecutionResult;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuilder;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jfrog.maven.annomojo.annotations.MojoComponent;
import org.jfrog.maven.annomojo.annotations.MojoParameter;

import java.io.IOException;
import java.util.List;
import java.util.Properties;

/**
 * An abstract class that provides the capability to redirect a mojo such that
 * it will always be executed from a specific place in a project. The most common
 * use case is, in a subclass's execute function: <br />
 * <br />
 * <code><pre>{@code
 * public void execute () {
 *      if ( executeWithNewPath("someGroupId","someArtifactId","this-goal-name") ) {
 *          return;
 *      }
 *      //Normal code here...
 * }
 * }</pre></code>
 * <br />
 *
 * {@link ARunAnywhereMojo#executeWithNewPath(String, String, String)} will return <code>true</code>
 * if it has had to relocate, and <code>false</code> otherwise, so if the mojo is
 * initially run from the wrong place, <code>executeWithNewPath</code> will launch
 * a new invocation of the mojo in the correct location and return <code>true</code>, at which
 * point the original invocation need do no further work, the sub-invocation will
 * have already done all the real work. <br />
 *
 * In this case, the correct location will be determined by examining the entire maven
 * project hierarchy that the originally used project is located in and looking for
 * a module with the specified group id and artifact id. Once found, the run anywhere
 * mojo will use the base directory of that module as the launching point. If it can't
 * find a matching module, the mojo will fail.
 *
 * @author Thor - nanderson@thetus.com
 */

public abstract class ARunAnywhereMojo extends AbstractMojo {

    @MojoParameter(required = true, expression = "${project}", readonly = true,
            description = "Builtin maven parameter")
    protected MavenProject project;

    @MojoComponent(description = "Built in maven component")
    protected ProjectBuilder projectBuilder;

    @MojoParameter(expression = "${session}", readonly = true,
            description = "Builtin maven parameter")
    protected MavenSession session;

    @MojoComponent(description = "Built in maven component")
    protected Maven maven;

    /**
     * Checks to see if the current project has the specified groupId and artifactId, if so, returns
     * false and performs no other actions. If not, looks for a module with that groupId/artifactId
     * pair, if it finds one, launches a new maven execution from that module with the specified
     * goal. If not, calls {@link ARunAnywhereMojo#getAlternateExecutionPath()},
     * if it returns non-null, launches a new maven execution from the returned path with the
     * specified goal. If not, throws an exception.
     *
     * @param desiredGroupId the groupId to re-launch from if not already present
     * @param desiredArtifactId the artifactId to re-launch from if not already present
     * @param goal the goal to re-launch if not already in correct location
     * @return <code>true</code> if relaunched in a different location, <code>false</code> if
     * already in the correct location
     * @throws MojoExecutionException if any issues are encountered
     */
    protected boolean executeWithNewPath(@NotNull String desiredGroupId, @NotNull String desiredArtifactId,
                                         @NotNull String goal) throws MojoExecutionException {
        if (!(project.getGroupId().equals(desiredGroupId) && project.getArtifactId().equals(desiredArtifactId))) {
            MavenProject desiredProject = null;

            if (session.getRequest().isProjectPresent()) {
                IMavenProjectHierarchyBuilder builder = lcl_initializeHierarchyBuilder();
                MavenProjectHierarchy hierarchy = builder.buildHierarchy(project, true, true);
                desiredProject = hierarchy.getProjectWithCoordinate(desiredGroupId, desiredArtifactId);
            }

            MavenExecutor executor = new MavenExecutor(session, maven);
            if (desiredProject == null) {
                String alternatePath = getAlternateExecutionPath();
                if (alternatePath != null) {
                    executor.setBaseProject(alternatePath);
                } else {
                    throw new MojoExecutionException("No project found");
                }
            } else {
                executor.setBaseProject(desiredProject);
            }

            String fullGoalName = findGoal(goal);
            if (fullGoalName != null) {
                executor.addGoal(fullGoalName);
                MavenExecutionResult executionResult = executor.execute();
                lcl_handleExecutionResult(executionResult);
            } else {
                throw new MojoExecutionException("Could not find a goal from the current session that matches the "
                        + "provided goal \"" + goal + "\"");
            }
            return true;
        }
        return false;
    }

    protected void executeWithNewPathByProperty(@NotNull String groupIdProperty, @NotNull String artifactIdProperty,
                                                @NotNull String fullGoal) throws MojoExecutionException {
        IMavenProjectHierarchyBuilder builder = lcl_initializeHierarchyBuilder();
        MavenProjectHierarchy hierarchy = builder.buildHierarchy(project, true, true);
        Properties props = hierarchy.getTopLevelProject().getProperties();
        if (!props.containsKey(groupIdProperty) || !props.containsKey(artifactIdProperty)) {
            throw new MojoExecutionException("The properties to the RunAnywhereMojo the indicate the groupId and artifactId of the Mojo to run were not specified.");
        }
        String groupId = props.getProperty(groupIdProperty);
        String artifactId = props.getProperty(artifactIdProperty);

        MavenProject desiredProject = hierarchy.getProjectWithCoordinate(groupId, artifactId);
        if (desiredProject == null) {
            throw new MojoExecutionException("No project found");
        }

        MavenExecutor executor = new MavenExecutor(session, maven);
        executor.setBaseProject(desiredProject);
        executor.addGoal(fullGoal);
        MavenExecutionResult result = executor.execute();
        lcl_handleExecutionResult(result);
    }

    /**
     * Convenience function for processing the results of a {@link MavenExecutor} execution. In the
     * case of the {@link ARunAnywhereMojo} we want to provide a somewhat helpful guess if the user
     * encountered an {@link IOException} (since the run anywhere redirection plays havoc with
     * relative paths), and propagate any exceptions encountered up.
     *
     * @param executionResult the result of the excution
     * @throws MojoExecutionException if an exception was encountered while performing the re-launched
     * execution
     */
    private void lcl_handleExecutionResult(MavenExecutionResult executionResult) throws MojoExecutionException {
        if (executionResult.hasExceptions()) {
            List<Throwable> executorExceptions = executionResult.getExceptions();
            for (Throwable t : executorExceptions) {
                getLog().error(t);
                for (Throwable c = t.getCause(); c != null; c = c.getCause()) {
                    if (c instanceof IOException) {
                        getLog().warn("#######################################################");
                        getLog().warn("IOExceptions may be caused by the MavenExecutor used by ARunAnywhereMojo");
                        getLog().warn("not being able to handle relative paths in your maven plugin.");
                        getLog().warn("Please replace any relative paths (./pathname) with absolute paths.");
                        getLog().warn("#######################################################");
                    }
                }
            }
            throw new MojoExecutionException(executionResult.getExceptions().get(0).toString());
        }
    }

    /**
     * Finds the full goal name that best matches the provided goal fragment. Works by
     * simply examining all of to the goals for the current maven session and finding
     * the one that ends with the provided fragment.
     *
     * @param goalFragment the goal fragment to match
     * @return the full name of the goal that matches the fragment or <code>null</code> if
     * one could not be found
     */
    protected @Nullable String findGoal(final String goalFragment) {
        List<String> currentGoals = session.getRequest().getGoals();
        String fullGoalName = null;
        for (String potentialGoal : currentGoals) {
            if (potentialGoal.endsWith(goalFragment)) {
                fullGoalName = potentialGoal;
            }
        }
        return fullGoalName;
    }

    /**
     * If the module specified in {@link ARunAnywhereMojo#executeWithNewPath(String, String, String)} cannot
     * be found in the current project tree, this function is called and if the return value is non-null will attempt
     * to use the returned path as the base of the new execution. By default, this function will return <code>null</code>
     * and must be overridden in a subclass in order return anything else.
     * @return the execution path to use if the requested groupId:artifactId is not found
     */
    protected String getAlternateExecutionPath() {
        try{
            return lcl_getAlternateExecutionPath();
        } catch (MojoExecutionException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            return null;
        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            return null;
        }
    }

    /**
     * Here temporarily for compatability reasons
     * @return
     */
    protected String lcl_getAlternateExecutionPath() throws MojoExecutionException , IOException {
        return null;
    }

    /**
     * Overridable hook for creating a hierarchy builder which will be used to create the
     * hierarchy we analyze to look for new places to execute
     * @return an ojbect capable of building a project hierarchy
     */
    protected IMavenProjectHierarchyBuilder lcl_initializeHierarchyBuilder () {
        return new DefaultMavenProjectHierarchyBuilder(projectBuilder, session, getLog());
    }
}
