/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.execution;

import com.thetus.oss.maven.util.RelativeFilenameUtils;
import org.apache.maven.Maven;
import org.apache.maven.execution.MavenExecutionRequest;
import org.apache.maven.execution.MavenExecutionResult;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.util.*;

/**
 * <p>
 * Allows you to launch maven lifecycles and plugins from within a maven run. This class requires
 * access to a {@link MavenSession} and {@link Maven} object, the first of which can be acquired as a
 * mojo parameter with the expression ${session}, the second as a mojo component.
 *</p>
 * <p>
 * Be aware, the MavenExecutor does not fork a new JVM. As a result, the working directory and other JVM
 * state will be the same as they are in the process using the executor. Mojos that are called directly
 * or indirectly by the executor must not rely on relative paths for files, since those paths might well
 * be resolved relative to a different working directory than expected. Where possible, using a
 * {@link MavenProject}'s <code>getBaseDir()</code> function explicitly to create an absolute path would be
 * preferable.
 *</p>
 * <p>
 * While not absolutely guaranteed (the internal workings of maven are quite complex), the state of the
 * maven execution environment after a call to <code>execute()</code> <i>should</i> be the same as it was before that call.
 * That is, for the launching maven environment, the goals executed, which projects are enabled,
 * what properties are set to, and so on, should not in any way be affected by the {@link MavenExecutor}.
 * </p>
 * <p>
 *     <b>Example:</b><br />
 *     <code><pre>{@code
 *     public void execute () {
 *      //Run a "mvn validate" on the project whose pom is located two up from the
 *      //current project's directory
 *
 *      MavenExecutor executor = new MavenExecutor(m_session,m_maven);
 *      File someOtherProjectDir = new File(m_project.getBasedir(),"../../");
 *      executor.setBaseProject(someOtherProjectDir.getAbsolutePath());
 *      executor.addGoal("validate");
 *
 *      MavenExecutionResult result = executor.execute();
 *      if ( result.hasExceptions() ) {
 *          getLog().error("Exceptions encountered while trying to validate");
 *      }
 *     }
 *     }</pre></code>
 *
 * </p>
 *
 *
 * @author Samson "Zigurd" Bock
 */
public class MavenExecutor {
    enum ProjectSetState { Unset , Empty , NonEmpty };
    MavenSession m_session;
    Maven m_maven;

    List<String> m_goals = new ArrayList<String>();
    List<String> m_projects = new ArrayList<String>();
    Map<String,String> m_properties = new HashMap<String,String>();
    boolean m_buildDownstream;
    boolean m_buildUpstream;
    MavenProject m_baseProject = null;
    String m_baseProjectPath = null;
    ProjectSetState m_projectSet = ProjectSetState.Unset;

    /**
     * Creates a MavenExecutor from the provided builtin maven objects. When initially created, the maven executor
     * is in the same state as the mojo that created it. That is, if one were to call {@link MavenExecutor#execute()}
     * immediately after construction, a new build identical to the one that had been originally launched would
     * start. This would create an infinite loop, of course, and is not advisable. It is simply useful for illustrating
     * that any modifications made to the way the executor will execute are working from the base state of the launching
     * mojo.
     * @param m_session the builtin maven Session representation,
     * can be acquired by use of the {@link org.jfrog.maven.annomojo.annotations.MojoParameter} annotation with the
     * <code>${session}</code> expression
     * @param m_maven the builtin maven self representation, can be acquired by use of the
     * {@link org.jfrog.maven.annomojo.annotations.MojoComponent} annotation
     */
    public MavenExecutor(@NotNull MavenSession m_session, @NotNull Maven m_maven) {
        this.m_session = m_session;
        this.m_maven = m_maven;

        String makeBehavior = m_session.getRequest().getMakeBehavior();
        m_buildDownstream = makeBehavior != null &&
                (makeBehavior.equals(MavenExecutionRequest.REACTOR_MAKE_DOWNSTREAM) ||
                        makeBehavior.equals(MavenExecutionRequest.REACTOR_MAKE_BOTH));
        m_buildUpstream = makeBehavior != null &&
                (makeBehavior.equals(MavenExecutionRequest.REACTOR_MAKE_UPSTREAM) ||
                        makeBehavior.equals(MavenExecutionRequest.REACTOR_MAKE_BOTH));
        m_baseProject = m_session.getCurrentProject();
    }

    /**
     * Adds a goal to the desired list of goals to execute. Goals will be executed in the order that they are
     * added with the same logic as if they had been sequentially appended to a command line maven command.
     * If this function is never called, the goals executed will be the same as the list of goals that the
     * command that launched this mojo had.
     * @param goalName the name of the goal to use (verify, test, install, javadoc:jar, etc...)
     */
    public void addGoal(String goalName) {
        m_goals.add(goalName);
    }

    /**
     * Adds a project to the list of selected projects, as designated by its path relative to the root project.
     * If this function (or the other addProject(s) variants) are never called, all projects in the base project's
     * reactor will be used.
     *
     * @param projectRelativePath the path, relative to the base project
     */
    public void addProject(String projectRelativePath) {
        m_projects.add(projectRelativePath);
        m_projectSet = ProjectSetState.NonEmpty;
    }

    /**
     * Sets the base project to execute from. In terms of equivalence to a command line build, this would be the
     * pom that you launched the maven command from. This function should be called prior to any of the addProject(s)
     * variants because it changes the directory that relative paths are relative to.
     *
     * @param baseProject the project to execute from
     */
    public void setBaseProject(@NotNull MavenProject baseProject) {
        this.m_baseProject = baseProject;
        this.m_baseProjectPath = null;
    }

    /**
     * Sets the base project path to execute from. In terms of equivalence to a command line build, this would be the
     * pom that you launched the maven command from. This function should be called prior to any of the addProject(s)
     * variants because it changes the directory that relative paths are relative to.
     *
     * @param baseProjectPath the path to the base project
     */
    public void setBaseProject(@NotNull String baseProjectPath) {
        this.m_baseProjectPath = baseProjectPath;
        this.m_baseProject = null;
    }

    /**
     * Adds a project to the list of selected projects.
     * If this function (or the other addProject(s) variants) are never called, all projects in the base project's
     * reactor will be used.
     *
     * @param project the path, relative to the base project
     */
    public void addProject(@NotNull MavenProject project) {
        MavenProject curP = m_baseProject;

        String baseURI = curP.getBasedir().getAbsolutePath();
        String projectURI = project.getBasedir().getAbsolutePath();

        String relativeURI = RelativeFilenameUtils.getRelativePath(projectURI, baseURI);
        assert relativeURI != null : "The maven executor encountered a null relative URI, you may have passed in an invalid" +
                                     " project or project path\n" +
                                     "Base URI: " + baseURI + "\nProject URI: " + projectURI;
        addProject(relativeURI.toString());
    }

    /**
     * Adds a collection of projects to the list of selected projects.
     * If this function (or the other addProject(s) variants) are never called, all projects in the base project's
     * reactor will be used.
     *
     * @param projects the path, relative to the base project
     */
    public void addProjects(Collection<MavenProject> projects) {
        //we want to behave a bit differently if someone says "use these projects" and there are none of them
        //however, we don't want to use this different behavior if they have already supplied other projects
        //to use.
        if ( projects.isEmpty() && m_projectSet == ProjectSetState.Unset ) { m_projectSet = ProjectSetState.Empty; }
        for (MavenProject mp : projects) {
            addProject(mp);
        }
    }

    /**
     * Sets whether or not upstream (dependency) projects should be built, in addition to selected projects. N.B. Projects
     * not present in the reactor will never be built, even if they are dependencies and this flag is turned on. Flag is
     * disabled by default.
     * @param truth true if dependencies should be built as well as selected projects, false otherwise
     */
    public void setBuildUpstream(boolean truth) {
        m_buildUpstream = truth;
    }

    /**
     * Sets whether or not downstream (dependent) projects should be built, in addition to selected projects. N.B. Projects
     * not present in the reactor will never be built, even if they are dependent and this flag is turned on. Flag is
     * disabled by default.
     * @param truth true if dependents should be built as well as selected projects, false otherwise
     */
    public void setBuildDownstream(boolean truth) {
        m_buildDownstream = truth;
    }

    /**
     * Indicates that given key/value pair should be set when the {@link MavenExecutor#execute()} is called. This
     * should be functionally equivalent to passing <code>-DpropertyKey=propertyValue</code> on the command line.
     * @param propertyKey key to of the property
     * @param propertyValue value to set it to
     */
    public void setProperty (String propertyKey, String propertyValue) {
        m_properties.put(propertyKey,propertyValue);
    }

    /**
     * Perform the maven execution using the currently configured settings of the executor. Launches a subordinate,
     * self-contained maven execution using the supplied selected projects, goals, properties, etc. Upon completion
     * the state of maven will be set back to what it was prior to execution, and the result of the execution will
     * be returned for processing. It is the responsibility of the caller to determine whether or not the build
     * succeeded, and what to do in case of failure in terms of error reporting and handling. Failure of the subordinate
     * execution will not cause the launching execution to fail.
     *
     * @return the result of the execution, contains information on any exceptions encountered
     */
    public MavenExecutionResult execute() {
        MavenExecutionRequest request = m_session.getRequest();
        List<String> oldGoals = request.getGoals();
        List<String> oldProjects = request.getSelectedProjects();
        String oldMakeBehavior = request.getMakeBehavior();
        String oldBaseDir = request.getBaseDirectory();
        File oldPom = request.getPom();

        //Set all the properties we want, keeping copies of their original values
        Map<String,String> oldProperties = new HashMap<String,String>();
        for ( Map.Entry<String,String> entry : m_properties.entrySet() ){
            oldProperties.put(entry.getKey(),m_session.getRequest().getSystemProperties().getProperty(entry.getKey()));
            m_session.getRequest().getSystemProperties().setProperty(entry.getKey(),entry.getValue());
        }

        request.setGoals(m_goals);
        if ( m_projectSet == ProjectSetState.Empty ){
            request.setSelectedProjects(Collections.<String>emptyList());
            System.out.println("Maven Executor received and empty set of projects to run, nothing will be run");
        }
        else if ( m_projects != null && m_projects.size() > 0) {
            request.setSelectedProjects(m_projects);
        }
        
        String makeBehavior = request.getMakeBehavior();
        if (m_buildDownstream && m_buildUpstream) {
            makeBehavior = MavenExecutionRequest.REACTOR_MAKE_BOTH;
        } else if (m_buildDownstream) {
            makeBehavior = MavenExecutionRequest.REACTOR_MAKE_DOWNSTREAM;
        } else if (m_buildUpstream) {
            makeBehavior = MavenExecutionRequest.REACTOR_MAKE_UPSTREAM;
        }
        request.setMakeBehavior(makeBehavior);

        if (m_baseProjectPath == null && m_baseProject != null) {
            request.setBaseDirectory(m_baseProject.getBasedir());
            request.setPom(m_baseProject.getFile());
        } else if (m_baseProject == null && m_baseProjectPath != null) {
            request.setBaseDirectory(new File(m_baseProjectPath));
            request.setPom(new File(m_baseProjectPath, "pom.xml"));
        }


        //===========================================================
        MavenThreadExecutor mavenThreadExecutor = new MavenThreadExecutor(m_maven, request);
        Thread thread = new Thread(mavenThreadExecutor);
        thread.start();
        try {
            thread.join();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        MavenExecutionResult result = mavenThreadExecutor.getResult();
        //===========================================================

        //Reset state of the maven environment to what it was before execution
        for ( Map.Entry<String,String> entry : oldProperties.entrySet() ){
            String k = entry.getKey();
            String v = entry.getValue();
            //Properties objects don't like it when you try to use a null
            if ( k != null && v != null ){
                m_session.getRequest().getSystemProperties().put(k,v);
            }
            else if ( k != null ){
                m_session.getRequest().getSystemProperties().remove(k);
            }
        }

        request.setGoals(oldGoals);
        request.setSelectedProjects(oldProjects);
        request.setMakeBehavior(oldMakeBehavior);
        request.setBaseDirectory(new File(oldBaseDir));
        request.setPom(oldPom);

        return result;
    }

    /**
     * Runs execute in a new thread to prevent issues with the classloader.
     */
    protected class MavenThreadExecutor implements Runnable {

        Maven m_maven;
        MavenExecutionRequest m_request;
        MavenExecutionResult m_result;

        public MavenThreadExecutor (Maven maven, MavenExecutionRequest request) {
            m_maven = maven;
            m_request = request;
        }

        public void run() {
             m_result = m_maven.execute(m_request);
        }

        public MavenExecutionResult getResult() {
            return m_result;
        }
    }

    public MavenExecutionResult execute( boolean handleResult ) throws MojoExecutionException {
        MavenExecutionResult res = execute();
        if ( handleResult ){
            System.out.println("### Execution Failed ###");
            if ( res.hasExceptions() ){
                for ( Throwable t : res.getExceptions() ){
                    t.printStackTrace();
                }
                throw new MojoExecutionException("sub-execution failed",res.getExceptions().get(0));
            }
        }
        return res;
    }
}
