/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.hierarchy;

import com.thetus.oss.maven.dependency.IDependencyResolver;
import com.thetus.oss.maven.dependency.ILocalBinariesSource;
import com.thetus.oss.maven.dependency.MavenDependencyResolver;
import com.thetus.oss.maven.util.ArtifactSorter;
import com.thetus.oss.maven.util.DependencySorter;
import com.thetus.oss.maven.util.MavenProjectUtils;
import com.thetus.oss.maven.util.RelativeFilenameUtils;
import com.thetus.oss.maven.dependency.IDependencyResolver;
import com.thetus.oss.maven.dependency.ILocalBinariesSource;
import com.thetus.oss.maven.dependency.MavenDependencyResolver;
import com.thetus.oss.maven.util.ArtifactSorter;
import com.thetus.oss.maven.util.DependencySorter;
import com.thetus.oss.maven.util.MavenProjectUtils;
import com.thetus.oss.maven.util.RelativeFilenameUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.handler.ArtifactHandler;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.Model;
import org.apache.maven.model.Profile;
import org.apache.maven.model.building.ModelBuildingRequest;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuilder;
import org.apache.maven.project.ProjectBuildingException;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;

/**
 * A utility class for usage in maven mojos, can dynamically construct a maven project hierarchy (mostly equivalent
 * to a maven reactor) give a few built in maven components and a starting point. Can be used to discover a larger
 * hierarchy that a project is part of by examining the directory and pom structure. Once constructed, the project
 * hierarchy can be used to perform actions on all its constituent projects, find projects by coordinates, etc.
 *
 * The two built-in maven objects that a maven project hierarchy requires, a {@link ProjectBuilder} and a {@link MavenSession}
 * can be acquired by means of the <code>MojoComponent</code> annotation for the former, and the <code>MojoParameter</code>
 * annotation with <code>expression="${session}"</code>for the latter. This will, of course, only work inside a
 * maven mojo.
 *
 * @author Samson "Zigurd" Bock - sbock@thetus.com
 */
public class MavenProjectHierarchy implements ILocalBinariesSource {

    /**
     * A simple internal helper class that represents a query for a project in the project hierarchy. Currently, it only
     * supports searching by groupId, artifactId, version, or some combination of the three, with the exception that
     * searching by version only is considered invalid.
     */
    public class ProjectQuery {
        String m_groupId = null;
        String m_artifactId = null;
        String m_version = null;

        /**
         * Constructs a project query with the given criteria, any argument that is left null will match
         * anything, all non-null arguments will be matched exactly, excepting leading and trailing
         * whitespace.
         *
         * @param groupId    intended project's group ID to match, may be null to match any groupId
         * @param artifactId intended project's artifact ID to match, may be null to match any artifactId
         * @param version    intended project's version to match, may be null to match any version
         */
        ProjectQuery(@Nullable String groupId, @Nullable String artifactId, @Nullable String version) {
            m_groupId = (groupId == null ? groupId : groupId.trim());
            m_artifactId = (artifactId == null ? artifactId : artifactId.trim());
            m_version = (version == null ? version : version.trim());
        }

        /**
         * @return whether the query is valid, and can be used in a search. A query will be valid if at least one of either
         *         it's artifactId or groupId has been specified and non-null, or both.
         */
        protected boolean lcl_valid() {
            return (m_groupId != null || m_artifactId != null); //since searching by version alone doesn't really make since
            //one of the other two needs to be non-null
        }
    }

    /**
     * Simple internal node class that represents a project and its relation to its children.
     */
    public class Node {
        /**
         * Constructs a node with the {@link MavenProject} it wraps, when constructed a node has no children.
         *
         * @param m_project the project to wrap
         */
        protected Node(MavenProject m_project) {
            this.m_project = m_project;
        }

        /**
         * Gets the {@link MavenProject} wrapped by this node
         *
         * @return The {@link MavenProject} wrapped by this node
         */
        public MavenProject getProject() {
            return m_project;
        }

        /**
         * Gets the list of children this node has, representing subprojects in a hierarchy
         *
         * @return The list of Nodes hierarchically under this node, each child should contain a maven projec that is
         *         a subproject of the project wrapped by this node.
         */
        public List<Node> getChildren() {
            return m_children;
        }

        /**
         * Adds a child node to this node
         *
         * @param child the node to add
         */
        protected void lcl_addChild(Node child) {
            m_children.add(child);
        }

        protected MavenProject m_project = null;
        protected List<Node> m_children = new ArrayList<Node>();
    }

    /**
     * The initial project that was passed in to create the hierarchy. Will not, in the general case, be the root of
     * the hierarchy. It is preserved internally for use in the calls that actually build the hierarchy, for
     * convenience. It is also kept because often the original project is a subject of interest to the overall mojo
     * even if it is not terribly important in the context of the hierarchy.
     */
    protected MavenProject m_initialProject;

    /**
     * The Project Builder that is used to construct {@link MavenProject} objects from pom.xml files. Needs to be passed
     * in, at present. In theory, this should probably be replaced at some point in the future with a
     * {@link org.apache.maven.project.ProjectBuilder} object, which seems to be the new, shiny way to convert pom.xml
     * files to Maven Projects. At present, however, it just seems to fail out for a variety of reasons, so we use the
     * older and more trustworthy one.
     */
    protected ProjectBuilder m_projectBuilder;

    /**
     * The maven session, contains much useful state information
     */
    protected MavenSession m_session;

    /**
     * Supplied by constructor
     */
    protected Log m_logger;

    /**
     * Root of the hierarchy, should not be <code>null</code> once the hierarchy building has been completed.
     */
    protected Node m_root = null;

    /**
     * Project representation of the root of the tree, exists only for convenience purposes
     */
    protected MavenProject m_topLevelProject = null;

    /**
     * Flag indicating whether the hierarchy should search up through the directory structure for a lower level
     * aggregator pom file
     */
    protected boolean m_searchBackwardForAggregator = true;

    /**
     * Flag indicating whether or not to look in extra, subclass-specific places for pom files
     */
    protected boolean m_findExtra = false;

    /**
     * Most common constructor in normal, single base-project usage. Performs initialization and root finding starting
     * from the specified project.
     *
     * @param builder      the project builder to use to construct projects from poms
     * @param session      the maven session object, supplies information about repositories and project building
     * @param project      the base project to start hierarchy search from
     * @param logger       the log to output information to, if null, a default log is instantiated
     * @param lookBackward controls whether or not to search backward for the root aggregator, or assume that the given project is the root
     * @param findExtra    controls whether extra, sub-class specific locations are searched for aggregator poms
     */
    public MavenProjectHierarchy(ProjectBuilder builder, @NotNull MavenSession session, @NotNull MavenProject project, @NotNull Log logger, boolean lookBackward, boolean findExtra) {
        m_findExtra = findExtra;
        m_logger = logger;
        m_session = session;
        m_searchBackwardForAggregator = lookBackward;
        m_projectBuilder = builder;
        m_initialProject = project;
        lcl_initialize();
    }

    /**
     * Constructs a hierarchy from the specified pom.xml
     * file, by constructing a {@link MavenProject} from it and using it as the base for a hierarchy search.
     *
     * @param builder the project builder to use to construct projects from poms
     * @param session the maven session object, supplies information about repositories and project building
     * @param pom     the location of the pom.xml file to use as the initial project
     * @param logger  the log to output information to, if null, a default log is instantiated   @throws ProjectBuildingException if the pom.xml cannot be used to create a valid project
     * @param lookBackward controls whether or not to search backward for the root aggregator, or assume that the given project is the root
     * @param findExtra    controls whether extra, sub-class specific locations are searched for aggregator poms
     * @throws org.apache.maven.project.ProjectBuildingException
     *          when a project cannot be created from the given pom file
     */
    public MavenProjectHierarchy(ProjectBuilder builder, MavenSession session, File pom, Log logger, boolean lookBackward , boolean findExtra ) throws ProjectBuildingException {
        m_findExtra = findExtra;
        m_searchBackwardForAggregator = lookBackward;
        m_logger = logger;
        m_projectBuilder = builder;
        m_session = session;
        m_initialProject = lcl_createProjectFromPom(pom);
        lcl_initialize();
    }


    /**
     * Walks down to find the root node, then walks up from there to construct a full representation of the hierarchy.
     */
    protected void lcl_initialize() {
        m_root = new Node(getTopLevelProject());
        lcl_buildTree(m_root);
    }

    /**
     * Convenience method that prints the tree, starting from the root, in depth first fashion.
     */
    public void printTree() {
        printTree(m_root);
    }

    /**
     * Publicly accessible function for printing the hierarchy below a certain node
     *
     * @param n the node at which to start
     */
    public void printTree(Node n) {
        if (n != null) {
            lcl_printTree(n, "\t");
        }
    }


    /**
     * Convenience method that gathers together
     * all direct dependencies of the hierarchy.
     *
     * @return a set of all direct dependencies
     */
    public Set<Dependency> getDirectDependencies() {
        final Set<Dependency> dependencySet = createEmptyDependencySet();
        final List<MavenProject> projects = getProjectList();
        for (MavenProject project : projects) {
            MavenProject p = project;
            while (p != null) {
                dependencySet.addAll(p.getDependencies());
                p = p.getParent();
            }
        }
        return dependencySet;
    }

    /**
     * Simple helper method for creating {@link Set} objects that hold {@link Dependency} and which sort based on the
     * groupId:artifactId:version of the dependency.
     *
     * @return an empty set using the {@link com.thetus.oss.maven.util.DependencySorter} as a comparator
     */
    public static Set<Dependency> createEmptyDependencySet() {
        return new TreeSet<Dependency>(new DependencySorter());
    }

    /**
     * Simple helper method for creating {@link Set} objects that hold {@link Artifact} objects and which sort based on
     * the groupId:artifactId:version of the artifact.
     *
     * @return an empty set using the {@link com.thetus.oss.maven.util.ArtifactSorter} as a comparator
     */
    public static Set<Artifact> createEmptyArtifactSet() {
        return new TreeSet<Artifact>(new ArtifactSorter());
    }

    /**
     * Returns a list of all projects in the hierarchy, not guaranteed to be in any particular order
     *
     * @return A List of all Maven Projects in the hierarchy, not necessarily in any particular order
     */
    public List<MavenProject> getProjectList() {
        final List<MavenProject> projects = new ArrayList<MavenProject>();
        if (m_root != null) {
            lcl_buildUpProjectListDepthFirst(m_root, projects);
        }
        return projects;
    }

    /**
     * Helper function to recursively populate project list
     *
     * @param n   current node
     * @param ret list to put results in
     */
    protected void lcl_buildUpProjectListDepthFirst(Node n, List<MavenProject> ret) {
        ret.add(n.getProject());
        for (Node child : n.getChildren()) {
            lcl_buildUpProjectListDepthFirst(child, ret);
        }
    }

    /**
     * Returns a set of all the dependencies for the project hierarchy, represented as {@link Artifact} objects. Because
     * this function is wrapping some {@link Dependency} objects without any particularly sophisticated {@link ArtifactHandler}
     * the artifacts returned by this function may not be entirely complete. For example, if used with the <code>pathOf</code>
     * function of the {@link org.apache.maven.artifact.repository.ArtifactRepository} class, the returned path will
     * not have an extension suffix.
     *
     * @return a {@link Set} containing all dependencies of all projects in the hierarchy, represented as {@link Artifact}
     *         objects
     */
    public Set<Artifact> getAllDependencyArtifacts() {
        IDependencyResolver resolver = new MavenDependencyResolver(m_projectBuilder, m_session);
        Set<Artifact> artifacts = resolver.resolveDependencyArtifacts(this);
        return artifacts;
    }

    /**
     * Internal, recursive print tree method. Acts in a depth-first manner.
     *
     * @param n    the currently active node
     * @param tabs the prefix string to print prior to the main info
     */
    protected void lcl_printTree(Node n, String tabs) {
        m_logger.info(tabs + (n.getChildren().size() > 0 ? "+" : "-") + n.getProject().getGroupId() + ":" + n.getProject().getArtifactId());
        for (Node subN : n.getChildren()) {
            lcl_printTree(subN, tabs + "\t");
        }
    }

    /**
     * Recurse out from the given node to build the tree representation of the hierarchy. Works by grabbing all sub
     * modules (if any) from the active node, and calling this function on them.
     *
     * @param n current node to build branch from
     */
    protected void lcl_buildTree(Node n) {
        MavenProject project = n.getProject();
        Set<String> modulesToExamine = lcl_getAllChildModules(project.getOriginalModel());
        for (String modulePath : modulesToExamine) {
            if (modulePath != null) {
                File baseDir = project.getBasedir();
                String absolutePath = baseDir.getAbsolutePath();
                if (!absolutePath.endsWith("/")) {
                    absolutePath = absolutePath + "/";
                }
                absolutePath = absolutePath + modulePath;
                final File pomFile;
                if (modulePath.contains(".xml")) {
                    pomFile = new File(absolutePath);
                } else {
                    pomFile = new File(absolutePath, "pom.xml");
                }
                m_logger.debug("Attempting to recurse to:" + pomFile);
                if (pomFile.exists()) {
                    MavenProject subProject = null;
                    try {
                        subProject = lcl_createProjectFromPom(pomFile);
                    } catch (ProjectBuildingException e) {
                        m_logger.warn("While constructing a project hierarchy, failed to build pom.xml file into a valid project");
                        m_logger.warn("[Pom File: " + pomFile.getAbsolutePath() + "]");
                        m_logger.warn(e);
                    }
                    if (subProject != null) {
                        Node subNode = new Node(subProject);
                        n.lcl_addChild(subNode);
                        lcl_buildTree(subNode);
                    }
                } else {
                    m_logger.warn("While constructing a project hierarchy, failed to find module pom.xml");
                    m_logger.warn("[Pom file: " + pomFile.getAbsolutePath() + "]");
                }
            }
        }
    }

    /**
     * Finds the topmost aggregator pom in this hierarchy. Used when constructing the initial hierarchy
     * representation. Internally, this function just walks backwards through the directory tree, looking for files
     * named pom.xml. Those it finds it examines, in order to determine if they aggregate the original project. If
     * they do, it marks them as the best top level aggregator candidate and moves on.
     *
     * @return the topmost aggregator project, or the original project, if there are no aggregators that reference it.
     */
    public MavenProject getTopLevelProject() {
        /*  We memoize the top level aggregator because it shouldn't change during the run of the program
            and there's no call to go walking the directory hierarchy without cause.
        */
        if (m_topLevelProject == null) {
            if (m_searchBackwardForAggregator == false) {
                m_topLevelProject = m_initialProject;
            } else {
                m_topLevelProject = lcl_findTopLevelProject();
            }
        }
        return m_topLevelProject;
    }

//    /**
//     * Internal function, walks the directory hierarchy to find the topmost aggregator that
//     * contains (transitively or directly) the initial project.
//     * @return the topmost aggregator or the initial project if none was found
//     */
//    protected @NotNull MavenProject lcl_findTopLevelProject() {
//        MavenProject project = m_initialProject;
//        m_logger.debug("Searching for aggregator root...");
//        try {
//            File parentDir = project.getBasedir().getParentFile();
//
//            while (parentDir != null && parentDir.isDirectory()) {
//                m_logger.debug("Checking to see if " + parentDir + " is an aggregator parent");
//                File parentDirPom = new File(parentDir, "pom.xml");
//                if (parentDirPom.exists() && parentDirPom.isFile()) {
//                    MavenProject parentProject = lcl_createProjectFromPom(parentDirPom);
//                    if (lcl_checkIfAggregatorIsParent(parentProject, project)) {
//                        m_logger.debug(parentDir + " is an aggregator parent");
//                        project = parentProject;
//                    } else {
//                        m_logger.debug(parentDir + " is not an aggregator parent");
//                    }
//                }
//
//                parentDir = parentDir.getParentFile();
//            }
//
//            m_logger.debug("Local aggregation root is " + project.getBasedir());
//        } catch (ProjectBuildingException e) {
//            m_logger.warn(e);
//        }
//
//        return project;
//    }

    /**
     * Internal function, walks the directory hierarchy to find the topmost aggregator that
     * contains (transitively or directly) the initial project.
     * @return the topmost aggregator or the initial project if none was found
     */
    protected @NotNull MavenProject lcl_findTopLevelProject() {
        MavenProject project = m_initialProject;
        m_logger.debug("Searching for aggregator root...");
        try {
            File parentDir;
            if (m_session.getRequest().isProjectPresent()) {
                parentDir = project.getBasedir().getParentFile();
            } else {
                parentDir = new File(m_session.getRequest().getBaseDirectory());
            }

            while (parentDir != null && parentDir.isDirectory()) {

                if (m_findExtra) {
                    MavenProject extra = lcl_checkExtraLocationsForParentPom(parentDir, project);
                    if (extra != null) {
                        project = extra;
                        break;
                    }
                }

                m_logger.debug("Checking to see if " + parentDir + " is an aggregator parent");
                File parentDirPom = new File(parentDir, "pom.xml");
                if (parentDirPom.exists() && parentDirPom.isFile()) {
                    MavenProject parentProject = lcl_createProjectFromPom(parentDirPom);
                    if (lcl_checkIfAggregatorIsParent(parentProject, project)) {
                        m_logger.debug(parentDir + " is an aggregator parent");
                        project = parentProject;
                    } else {
                        m_logger.debug(parentDir + " is not an aggregator parent");
                    }
                }

                parentDir = parentDir.getParentFile();
            }

            m_logger.debug("Local aggregation root is " + project.getBasedir());
        } catch (ProjectBuildingException e) {
            m_logger.warn(e);
        }

        return project;
    }

    /**
     * Performs auxilliary checks for a parent aggregator, can be overridden by
     * subclasses to provide an extra set of locations to search
     * @param parentDir the current directory being searched
     * @param project the current top level maven project
     * @return a new top level {@link MavenProject} or <code>null</code>, if no valid project was found,
     * will always be <code>null</code> unless overridden
     */
    @Nullable protected MavenProject lcl_checkExtraLocationsForParentPom(final File parentDir, MavenProject project) {
        return null;
    }


    /**
     * Checks to see if the given parentProject aggregates the given project
     *
     * @param parentProject potential aggregator
     * @param project       aggregatee
     * @return <code>true</code> if the parent aggregates the project, <code>false</code> otherwise
     */
    protected boolean lcl_checkIfAggregatorIsParent(MavenProject parentProject, MavenProject project) {
        try {
            URI relativePath = new URI(FilenameUtils.separatorsToUnix(RelativeFilenameUtils.getRelativePath(project.getBasedir().toString(), parentProject.getBasedir().toString())));

            Set<String> childModules = lcl_getAllChildModules(parentProject.getOriginalModel());
            boolean found = false;
            for (String str : childModules) {
                found = lcl_checkForModuleMatch(relativePath, str);
                if (found) {
                    break;
                }
            }

            return found;
        } catch (URISyntaxException e) {
            m_logger.warn(e);
        }
        return false;
    }


    /**
     * Helper function for creating a project given a pom.xml file
     *
     * @param pomFile file to use to construct a project
     * @return the project representation
     * @throws ProjectBuildingException if the pom.xml file was invalid or failed to parse
     */
    protected MavenProject lcl_createProjectFromPom(File pomFile) throws ProjectBuildingException {
        return m_projectBuilder.build(pomFile, m_session.getProjectBuildingRequest().setValidationLevel(ModelBuildingRequest.VALIDATION_LEVEL_MINIMAL)).getProject();
    }

    /**
     * Internal function for gathering all projects in the hierarchy that match a given query.
     *
     * @param node      current node to examine
     * @param query     criteria for matching
     * @param resultSet collection to store the resulting projects in
     */
    protected void lcl_gatherProjectsMatchingQuery(Node node, final ProjectQuery query, Set<MavenProject> resultSet) {
        MavenProject project = node.getProject();
        if ((query.m_groupId == null || query.m_groupId.equals(project.getGroupId())) &&
                (query.m_artifactId == null || query.m_artifactId.equals(project.getArtifactId())) &&
                (query.m_version == null || query.m_version.equals(project.getVersion()))
                ) {
            resultSet.add(project);
        }

        for (Node n : node.getChildren()) {
            lcl_gatherProjectsMatchingQuery(n, query, resultSet);
        }

    }

    /**
     * Internal function for getting all projects that match a query, in list form. Internally this function delegates
     * most work to {@link MavenProjectHierarchy#lcl_gatherProjectsMatchingQuery(MavenProjectHierarchy.Node, MavenProjectHierarchy.ProjectQuery, java.util.Set)}.
     *
     * @param query the criteria for matching
     * @return a list of maven projects that matched the query
     */
    public List<MavenProject> getProjectsMatchingQuery(ProjectQuery query) {
        Set<MavenProject> set = new HashSet<MavenProject>();

        Node root = m_root;
        lcl_gatherProjectsMatchingQuery(root, query, set);

        List<MavenProject> res = new ArrayList<MavenProject>();
        res.addAll(set);
        return res;
    }

    /**
     * Gets the maven project that has the given coordinate, if any exists in this hierarchy, if one exists, or null otherwise.
     *
     * @param groupId    the project's group id
     * @param artifactId the project's artifact id
     * @param versionID  the project's version
     * @return The maven project matching given coordinates, or null if none exists in this hierarchy
     */
    public MavenProject getProjectWithCoordinate(@NotNull String groupId, @NotNull String artifactId, @NotNull String versionID) {
        List<MavenProject> res = getProjectsMatchingQuery(new ProjectQuery(groupId, artifactId, versionID));
        if (res.size() == 0) {
            return null;
        }
        return res.get(0);
    }

    /**
     * Public facing function that will return a single project with the specified groupId and artifactId, or <code>
     * null</code> if none exists.
     *
     * @param groupId    the desired groupId to look for
     * @param artifactId the desired artifactId to look for
     * @return the project in this hierarchy with the given groupId and artifactId, or null if none such exists
     */
    public
    @Nullable
    MavenProject getProjectWithCoordinate(@NotNull String groupId, @NotNull String artifactId) {
        List<MavenProject> res = getProjectsMatchingQuery(new ProjectQuery(groupId, artifactId, null));
        if (res.size() == 0) {
            return null;
        }
        return res.get(0);
    }

    /**
     * @param groupId    the desired groupId to look for
     * @param artifactId the desired artifactId to look for
     * @return <code>true</code> if a project with the given coordinates exists in this hierarchy, <code>false</code> otherwise
     */
    public boolean containsProjectWithCoordinate(@NotNull String groupId, @NotNull String artifactId) {
        List<MavenProject> res = getProjectsMatchingQuery(new ProjectQuery(groupId, artifactId, null));
        return !res.isEmpty();
    }

    /**
     * Public facing method for acquiring a list of all projects in the hierarchy that have a given groupId
     *
     * @param groupId the groupId to look for
     * @return a list of maven projects with the given groupId
     */
    public
    @NotNull
    List<MavenProject> getProjectsWithGroupId(String groupId) {
        List<MavenProject> res = getProjectsMatchingQuery(new ProjectQuery(groupId, null, null));
        return res;
    }

    /**
     * Public facing method for acquiring a list of all projects in the hierarchy that have a given artifactId
     *
     * @param artifactId the artifactId to look for
     * @return a list of maven projects with the given artifactId
     */
    public
    @NotNull
    List<MavenProject> getProjectsWithArtifactId(String artifactId) {
        List<MavenProject> res = getProjectsMatchingQuery(new ProjectQuery(null, artifactId, null));
        return res;
    }

    /**
     * Convenience function wrapped around {@link MavenProjectHierarchy#getProjectsWithArtifactId(String)} that returns
     * only one result or null, if no results exist.
     *
     * @param artifactId the artifactId to look for
     * @return a maven project with the given artifactId, or <code>null</code> if none exists
     */
    public
    @Nullable
    MavenProject getProjectWithArtifactId(String artifactId) {
        List<MavenProject> res = getProjectsWithArtifactId(artifactId);
        if (res.size() == 1) {
            return res.get(0);
        } else if (res.size() > 1) {
            m_logger.warn("getProjectWithArtifactId() had more than one result, only returning the first");
            return res.get(0);
        } else {
            return null;
        }
    }

    /**
     * Convenience function for checking to see if a URI loosely matches a string, for the purpose of checking if a
     * module path refers to a project.
     *
     * @param relativePath the path to compare to
     * @param str          the string to compare against
     * @return <code>true</code> if the paths are loosely equivalent, false otherwise
     * @throws URISyntaxException if a malformed url is created as part of the module check, should never happen
     */
    private boolean lcl_checkForModuleMatch(URI relativePath, String str) throws URISyntaxException {
        String path = str;
        if (path.startsWith("./")) {
            path = path.substring(2);
        }

        return relativePath.equals(new URI(path));
    }

    /**
     * Convenience function for gathering all child modules of a given {@link Model}
     *
     * @param model the model to gather modules from
     * @return a set of all modules from the model
     */
    private
    @NotNull
    Set<String> lcl_getAllChildModules(Model model) {
        Set<String> childModules = new TreeSet<String>();
        childModules.addAll(model.getModules());
        for (Profile profile : model.getProfiles()) {
            childModules.addAll(profile.getModules());
        }

        return childModules;
    }

    /**
     * @param pluginArtifactId artifactId of the plugin to look for usages of
     * @param goalName         name of the goal to look for usages of
     * @return a list of all projects in the hierarchy that make use of the given goal of the plugin
     *         with the specified artifactId
     */
    public List<MavenProject> getProjectsWithPluginGoal(@NotNull String pluginArtifactId, @Nullable String goalName) {
        List<MavenProject> ret = new ArrayList<MavenProject>();
        List<MavenProject> src = getProjectList();
        for (MavenProject p : src) {
            if (MavenProjectUtils.doesProjectUsePluginGoal(p, pluginArtifactId, goalName)) {
                ret.add(p);
            }
        }
        return ret;
    }

    /**
     * Gets the local binary (generally a target/classes directory) for the given coordinates.
     *
     * @param groupId             artifact's group id to find
     * @param artifactId          artifact's artifact id to find
     * @param version             artifact's version to find
     * @param includeTestBinaries whether or not to include test classes/binaries in the returned set
     * @return a list of local binary paths, in string form
     */
    @Override
    public @Nullable List<String> getLocalBinaryDirectoriesForCoordinate(@NotNull String groupId, @NotNull String artifactId, @NotNull String version, boolean includeTestBinaries) {
        MavenProject project = getProjectWithCoordinate(groupId, artifactId, version);
        if (project != null) {
            List<String> ret = new ArrayList<String>();
            ret.add(project.getBuild().getOutputDirectory());
            if (includeTestBinaries) {
                ret.add(project.getBuild().getTestOutputDirectory());
            }
            return ret;
        } else {
            return null;
        }
    }
}
