/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.hierarchy;

import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Dependency;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuilder;
import org.jetbrains.annotations.Nullable;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.File;
import java.io.IOException;
import java.util.*;

/**
 * This class provides more detailed information and analysis of a {@link MavenProjectHierarchy} than
 * you can get from the {@link MavenProjectHierarchy} itself. It can provide a variety of useful
 * things: a list of the maven properties used to reference versions of other projects in the same
 * tree, the root poms that provide version numbers to their children, a topological sort of
 * all modules and of entire sub-trees, etc.
 *
 *
 * @author Samson "Zigurd" Bock - sbock@thetus.com
 */
public class MavenProjectHierarchyAnalysis {
    /**
     * Data container class that holds information about a dependency, along with the metadata
     * this class uses in its analysis
     */
    public static class DependencyItem {
        public ArtifactItem m_dependentArtifact;
        public ArtifactItem m_dependencyArtifact;
        public String m_groupId;
        public String m_artifactId;
        public String m_version;
        public boolean m_literalVersion;
    }

    /**
     * Data container class that holds information about an artifact, along with the metadata
     * this class uses in its analysis
     */
    public static class ArtifactItem {
        public String m_groupId;
        public String m_artifactId;
        public String m_version;
        public VersionSources m_versionType;
        public String m_parentGroupId;
        public String m_parentArtifactId;
        public List<DependencyItem> m_dependencies = new ArrayList<DependencyItem>();
        public List<ArtifactItem> m_dependents = new ArrayList<ArtifactItem>();
        //Flag for use with topological sorting
        public boolean m_visited = false;
    }

    /**
     * The different sources an artifact can have for its version number. "Parent" if none is specified
     * and it determines its version implicitly from its parent, "Property" if a maven property of some
     * kind e.g. ${version.prop} is used, and "Literal" if a string literal, e.g. "1.0-SNAPSHOT" is used.
     */
    public enum VersionSources { Parent, Property, Literal} ;

    /**
     * A list of the contents of all &lt;dependency&gt; tags in all poms in the maven project
     * that are not part of a dependency management section
     */
    protected List<DependencyItem> m_dependencyItems = new ArrayList<DependencyItem>();

    /**
     * A list of the contents of all &lt;dependency&gt; tags in all poms in the maven project
     * that are part of a dependency management section
     */
    protected List<DependencyItem> m_dependencyManagementItems = new ArrayList<DependencyItem>();

    /**
     * A list of all artifacts present in the hierarchy, represented in this class's internal format
     */
    protected List<ArtifactItem> m_artifactItems = new ArrayList<ArtifactItem>();

    /**
     * All artifacts in the project, indexed by their groupId:artifactId coordinate
     */
    protected Map<String, ArtifactItem> m_artifactItemsByCoordinate = new HashMap<String, ArtifactItem>();

    /**
     * A set of all the properties that are defined in the poms of this project
     */
    protected Set<String> m_definedProperties = new HashSet<String>();

    /**
     * Mapping of properties to the artifacts that define them
     */
    protected Map<String, List<ArtifactItem>> m_definedPropertiesSources = new HashMap<String, List<ArtifactItem>>();

    /**
     * All artifacts in the project, sorted in topological order based on their dependencies
     */
    protected List<ArtifactItem> m_topologicalSortedArtifacts = new ArrayList<ArtifactItem>();

    /**
     * Maven object used to acquire information about the current maven session, including project building
     * request settings. Acquired through use of the <code>MojoParameter</code> annotation with the
     * expression ${session}
     */
    protected MavenSession m_session;

    /**
     * The maven-toolbox representation of a maven project hierarchy.
     */
    protected MavenProjectHierarchy m_hierarchy;

    /**
     * The maven provided logger
     */
    protected Log m_log;

    /**
     * Builds a new analysis based on the provided hierarchy.
     * @param hierarchy the hierarchy to analyze
     * @param log built-in maven logger
     */
    public MavenProjectHierarchyAnalysis(MavenProjectHierarchy hierarchy, Log log) {
        m_hierarchy = hierarchy;
        m_log = log;

        lcl_analyze();
    }

    /**
     * Gets all the projects in the analyzed hierarchy that do not depend on anything else for
     * their version number. In practical terms, this is all artifacts that have a pom with a
     * personal &lt;version&gt; tag (that is, not a version tag referring to their parent or
     * some dependency).
     *
     * This list is useful when looking to update the versions of a project hierarchy, since
     * the projects that provide their own versions will be the ones that will need to be
     * changed.
     *
     * @return a list of all projects that provide their own version number
     */
    public List<MavenProject> getRootVersionedProjects() {
        List<MavenProject> ret = new ArrayList<MavenProject>();
        for (ArtifactItem item : m_artifactItems) {
            if (item.m_versionType == VersionSources.Literal || item.m_versionType == VersionSources.Property) {
                MavenProject project = m_hierarchy.getProjectWithCoordinate(item.m_groupId, item.m_artifactId);
                if (project != null) {
                    ret.add(project);
                } else {
                    m_log.debug("Root Versioned Project not found in hierarchy: " + item.m_groupId + ":" + item.m_artifactId + ":" + item.m_version);
                }
            }
        }
        return ret;
    }

    /**
     * Gets a list of all the properties defined in this hierarchy that are used to represent
     * the versions of other projects in the hierarchy. For example, if project <code>B</code>
     * has a dependency on another project in the hierarchy, <code>A</code>, and it defines the
     * version of <code>A</code> that
     * it depends on as <code>${thetus.versions.A}</code>, and in some pom in the project there
     * is a &lt;properties&gt;&lt;thetus.version.A&gt;1.0.0&lt;/thetus.version.A&gt;&lt;/properties&gt;
     * section, "thetus.version.A" would be among the values returned by this method.
     *
     * @return a list of all properties used to specify inter-project dependency versions
     */
    public List<String> getInterProjectVersionProperties() {
        List<String> ret = new ArrayList<String>();

        Set<String> properties = new HashSet<String>();
        List<DependencyItem> allDepItems = getAllDependencyItems();
        for (DependencyItem item : allDepItems) {
            if (item.m_version != null && !item.m_literalVersion) {
                for (ArtifactItem ai : m_artifactItems) {
                    if (ai.m_groupId.equalsIgnoreCase(item.m_groupId) && ai.m_artifactId.equalsIgnoreCase(item.m_artifactId)) {
                        properties.add(item.m_version.trim());
                        break;
                    }
                }
            }
        }

        for (String str : properties) {
            if (m_definedProperties.contains(str)) {
                ret.add(str);
            }
        }
        return ret;
    }

    /**
     * @return a list of all dependencies items in the analysis' internal format, including those
     * from both build/dependencies and build/dependencyManagement/dependencies sections.
     */
    public List<DependencyItem> getAllDependencyItems() {
        List<DependencyItem> allDepItems = new ArrayList<DependencyItem>();
        allDepItems.addAll(m_dependencyItems);
        allDepItems.addAll(m_dependencyManagementItems);
        return allDepItems;
    }

    /**
     * Finds the project from the list of root versioned projects ({@see MavenProjectHierarchyAnalysis#getRootVersionedProjects})
     * that is linked with the given property. By way of example, say project <code>B</code>
     * has a dependency on another project in the hierarchy, <code>A</code>, and it defines the
     * version of <code>A</code> that
     * it depends on as <code>${thetus.versions.A}</code>. <code>A</code> doesn't provide a
     * version of its own, but rather inherits it from its parent, <code>super-pom-A</code>.
     * Calling this function with "thetus.versions.A" as the argument would return the
     * <code>super-pom-A</code> project.
     *
     * Internally, this function simply looks through all of the dependency sections in the
     * hierarchy for those that make use of the given property, once it finds one it examines
     * the dependency artifact. If the artifact that is depended on supplies is own version,
     * that artifact is returned, if not, its parent is examined, and so on.
     *
     * @param property the property to search for
     * @return the root versioned project that corresponds to the supplied version property
     */
    public MavenProject getRootVersionedProjectForProperty(String property) {
        MavenProject ret = null;
        List<DependencyItem> allDepItems = getAllDependencyItems();
        for (DependencyItem item : allDepItems) {
            if (item.m_version != null && !item.m_literalVersion && item.m_version.equals(property.trim())) {
                ArtifactItem art = item.m_dependencyArtifact;
                if (art == null) {
                    m_log.error("NULL encountered: " + item.m_groupId + ":" + item.m_artifactId);
                }
                while (art != null) {
                    if (art.m_versionType == VersionSources.Literal || art.m_versionType == VersionSources.Property) {
                        return m_hierarchy.getProjectWithCoordinate(art.m_groupId, art.m_artifactId);
                    }
                    art = getArtifactItem(art.m_parentGroupId, art.m_parentArtifactId);
                }
            }
        }
        return ret;
    }

    /**
     * @param groupId the groupId of the artifact to get
     * @param artifactId the artifactId of the artifact to get
     * @return the analysis' internal representation of the artifact with the given groupId and artifactId or <code>null</code>
     * if no such artifact exists in the current hierarchy
     */
    public @Nullable ArtifactItem getArtifactItem(String groupId, String artifactId) {
        return m_artifactItemsByCoordinate.get(groupId + ":" + artifactId);
    }

    /**
     * Given a list of project hierarchies that are sub-hierarchies of the analyzed hierarchy,
     * returns a list containing those hierarchies, sorted topologically
     * according to their inter-project dependencies. Any hierarchies included that are not
     * part of the analyzed hierarchy will be included last in the returned list.
     * @param hierarchies the list of sub-hierarchies to sort
     * @return a new list containing the provided hierarchies sorted in topological order based on
     * dependency relationships.
     */
    public List<MavenProjectHierarchy> getSubHierarchyTopologicalSort(List<MavenProjectHierarchy> hierarchies) {
        List<MavenProjectHierarchy> ret = new ArrayList<MavenProjectHierarchy>();
        for (ArtifactItem item : m_topologicalSortedArtifacts) {
            for (MavenProjectHierarchy h : hierarchies) {
                if (!ret.contains(h)) {
                    if (h.getProjectWithCoordinate(item.m_groupId, item.m_artifactId) != null) {
                        ret.add(h);
                        if (ret.size() == hierarchies.size()) {
                            return ret;
                        } //early exit, we've found all the hierarchies
                    }
                }
            }
        }
        for ( MavenProjectHierarchy h : hierarchies ) {
            if ( ! ret.contains(h) ) {
                ret.add(h);
            }
        }
        return ret;
    }

    /**
     * Given a source hierarchy and a list of possible target hierarchies, returns a list of those target hierarchies
     * that are dependent upon the source hierarchy.
     * @param srcHierarchy the hierarchy to find dependents of
     * @param hierarchies the pool of hierarchies to look at for dependents
     * @return the list of project hierarchies that are dependent on the source hierarchy
     */
    public List<MavenProjectHierarchy> getDependentSubHierarchies(MavenProjectHierarchy srcHierarchy, List<MavenProjectHierarchy> hierarchies) {
        List<MavenProjectHierarchy> ret = new ArrayList<MavenProjectHierarchy>();
        for (ArtifactItem item : m_topologicalSortedArtifacts) {
            if (srcHierarchy.getProjectWithCoordinate(item.m_groupId, item.m_artifactId) != null) {
                for (ArtifactItem dependent : item.m_dependents) {
                    if (dependent == null) {

                    } else {
                        for (MavenProjectHierarchy h : hierarchies) {
                            if (h != srcHierarchy && !ret.contains(h)) {
                                if (h.getProjectWithCoordinate(dependent.m_groupId, dependent.m_artifactId) != null) {
                                    ret.add(h);
                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }
        return ret;
    }


    /**
     * @param property the property to find definitions of
     * @return a list of artifacts -- in the analysis' internal representation -- that define the supplied property
     * in their pom
     */
    public List<ArtifactItem> getArtifactsDefiningProperty(String property) {
        List<ArtifactItem> artifacts = m_definedPropertiesSources.get(property);
        return artifacts != null ? artifacts : Collections.<ArtifactItem>emptyList();
    }

    /**
     * @param property the property to find definitions of
     * @return a list of the maven projects in the analyzed hierarchy that define the supplied property in their pom
     */
    public List<MavenProject> getProjectsDefiningProperty(String property) {
        List<ArtifactItem> artifacts = getArtifactsDefiningProperty(property);
        List<MavenProject> projects = new ArrayList<MavenProject>();
        for ( ArtifactItem item : artifacts ) {
            MavenProject project = m_hierarchy.getProjectWithCoordinate(item.m_groupId, item.m_artifactId);
            if ( project != null ) {
                projects.add(project);
            }
        }
        return projects;
    }

    /**
     * Perform the analysis of the project hierarchy, most of the actual work is done by
     * {@link MavenProjectHierarchyAnalysis#lcl_processProject(org.apache.maven.project.MavenProject)}
     * and {@link MavenProjectHierarchyAnalysis#lcl_postProcessDependencies()}
     */
    protected void lcl_analyze() {
        for (MavenProject project : m_hierarchy.getProjectList()) {
            try {
                lcl_processProject(project);
            } catch (ParserConfigurationException e) {
                m_log.error(e);
            } catch (SAXException e) {
                m_log.error(e);
            } catch (IOException e) {
                m_log.error(e);
            }
        }
        lcl_postProcessDependencies();
    }

    /**
     * Perform the post processing stage of the analysis, connects dependents
     * with dependencies, creates the main topologically sorted list of
     * projects, etc...
     */
    protected void lcl_postProcessDependencies() {
        //Go through all the dependency items collected from dependencies and
        //dependency management sections and populate its dependents section
        for (DependencyItem dep : getAllDependencyItems()) {
            String coordinate = dep.m_groupId + ":" + dep.m_artifactId;
            if (m_artifactItemsByCoordinate.containsKey(coordinate)) {
                dep.m_dependencyArtifact = m_artifactItemsByCoordinate.get(coordinate);
                dep.m_dependencyArtifact.m_dependents.add(dep.m_dependentArtifact);
            }
        }

        //Create the starting set for topological sorting by finding those
        //artifact items that have no dependents themselves
        Set<ArtifactItem> startSet = new HashSet<ArtifactItem>();
        for (ArtifactItem item : m_artifactItems) {
            if (item.m_dependents.isEmpty()) {
                startSet.add(item);
            }
        }

        //Run the recursive topological sorting function
        for (ArtifactItem item : startSet) {
            lcl_topologicalVisit(item);
        }
    }

    /**
     * Recursively topologically sort the artifacts based on dependency relationships
     * @param item the current item to sort
     */
    protected void lcl_topologicalVisit(ArtifactItem item) {
        if (!item.m_visited) {
            item.m_visited = true;
            for (DependencyItem dep : item.m_dependencies) {
                if (dep.m_dependencyArtifact != null) {
                    lcl_topologicalVisit(dep.m_dependencyArtifact);
                }
            }
            m_topologicalSortedArtifacts.add(item);
        }
    }

    /**
     * Document builder factory to use for all parsing needs
     */
    DocumentBuilderFactory m_documentBuilderFactory = DocumentBuilderFactory.newInstance();

    /**
     * Process the given project, primarily by sifting through the raw XML of its pom file. We do
     * this instead of relying on Maven's {@link MavenProject} representation because we need to
     * know not just the final effective project, but also the literal base representation. We need
     * to know not only the version of the project, but also how that version is defined, for
     * example.
     *
     * @param project the maven project to process
     * @throws ParserConfigurationException if anything goes wrong setting up the parser
     * @throws IOException if anything goes wrong reading in the information
     * @throws SAXException if anything goes wrong while parsing
     */
    protected void lcl_processProject(MavenProject project) throws ParserConfigurationException, IOException, SAXException {
        //We don't currently perform any analysis on aggregator poms
        if (!project.getModules().isEmpty()) {
            return;
        }

        DocumentBuilder db = m_documentBuilderFactory.newDocumentBuilder();
        File pom = project.getFile();

        Document doc = db.parse(pom);
        doc.getDocumentElement().normalize();

        Node projectNode = doc.getElementsByTagName("project").item(0);
        NodeList dependencyNodes = doc.getElementsByTagName("dependency");
        NodeList propertiesNodes = doc.getElementsByTagName("properties");
        NodeList propertyNodes = (propertiesNodes.getLength() > 0 ? propertiesNodes.item(0).getChildNodes() : null);

        //Create the internal representation for this project's artifact
        ArtifactItem art = new ArtifactItem();
        art.m_artifactId = project.getArtifactId();
        art.m_groupId = project.getGroupId();
        Node versionNode = lcl_getChildNodeWithTag(projectNode, "version");

        if (project.getParent() != null) {
            art.m_parentGroupId = project.getParent().getGroupId();
            art.m_parentArtifactId = project.getParent().getArtifactId();
        }

        if (versionNode == null) {
            art.m_versionType = VersionSources.Parent;
        } else {
            art.m_version = versionNode.getTextContent().trim();
            if (art.m_version.contains("${")) {
                art.m_versionType = VersionSources.Property;
            } else {
                art.m_versionType = VersionSources.Literal;
            }
        }

        m_artifactItems.add(art);
        m_artifactItemsByCoordinate.put(art.m_groupId + ":" + art.m_artifactId, art);

        //Loop through all the dependency nodes in order to populate the dependency item lists
        for (int i = 0; i < dependencyNodes.getLength(); ++i) {
            Node depNode = dependencyNodes.item(i);

            Node groupId = lcl_getChildNodeWithTag(depNode, "groupId");
            Node artifactId = lcl_getChildNodeWithTag(depNode, "artifactId");
            Node version = lcl_getChildNodeWithTag(depNode, "version");

            if ( artifactId == null || groupId == null ) {
                m_log.warn("Project encountered with an incomplete groupId/artifactId pair");
                m_log.warn("\tGroupId: " + (groupId == null ? groupId : groupId.getTextContent()));
                m_log.warn("\tArtifactId: " + (artifactId == null ? artifactId : artifactId.getTextContent()));
                m_log.warn("Skipping project");
                continue;
            }

            DependencyItem depItem = new DependencyItem();
            depItem.m_artifactId = artifactId.getTextContent().trim();
            depItem.m_groupId = groupId.getTextContent().trim();
            /*  If the groupId is defined by a property, we're going to rely on matching against the resolved
                MavenProject representation to find the actual groupId, we simply loop through all the dependencies
                in the MavenProject looking for the one that matches. If there are multiple <dependency> tags that
                have different groupIds but the same artifactId, and at least one of them defines the groupId by
                a property, this approach will fail, but we haven't seen that case as of yet.
             */
            if (depItem.m_groupId.contains("${")) {
                List<Dependency> dependencies = project.getDependencies();
                dependencies.addAll(project.getDependencyManagement().getDependencies());
                String foundGroupId = null;
                for (Dependency d : dependencies) {
                    if (d.getArtifactId().equalsIgnoreCase(depItem.m_artifactId)) {
                        if (foundGroupId != null && ! foundGroupId.equals(d.getGroupId())) {
                            m_log.warn("Multiple dependencies in project " + art.m_groupId + ":" + art.m_artifactId + " with the same artifactId, but different groupIds");
                        }
                        foundGroupId = d.getGroupId();
                    }
                }
                if (foundGroupId != null) {
                    depItem.m_groupId = foundGroupId;
                }
            }
            //If the groupId is still a property, then we haven't found a suitable replacement from the resolved
            //MavenProject, this can happen for legitimate reasons, but we should mention it anyway if we're debugging
            if (depItem.m_groupId.contains("${")) {
                m_log.debug("Couldn't resolve " + depItem.m_groupId + " in " + project.getFile());
            }

            depItem.m_version = version == null ? null : version.getTextContent().trim();
            depItem.m_literalVersion = version == null ? false : (!depItem.m_version.contains("${"));

            if (depNode.getParentNode().getParentNode().getNodeName().equalsIgnoreCase("dependencyManagement")) {
                m_dependencyManagementItems.add(depItem);
            } else {
                depItem.m_dependentArtifact = art;
                art.m_dependencies.add(depItem);
                m_dependencyItems.add(depItem);
            }
        }

        DependencyItem parentItem = new DependencyItem();
        parentItem.m_artifactId = project.getParent().getArtifactId();
        parentItem.m_groupId = project.getParent().getGroupId();
        parentItem.m_version = project.getParent().getVersion();
        parentItem.m_literalVersion = true;
        parentItem.m_dependentArtifact = art;
        art.m_dependencies.add(parentItem);
        m_dependencyItems.add(parentItem);

        //Loop through and record all of the properties that are defined in this pom
        if (propertyNodes != null) {
            for (int i = 0; i < propertyNodes.getLength(); ++i) {
                Node propNode = propertyNodes.item(i);
                String prop = "${" + propNode.getNodeName().trim() + "}";
                m_definedProperties.add(prop);
                List<ArtifactItem> artifacts = m_definedPropertiesSources.get(prop);
                if (artifacts == null) {
                    artifacts = new ArrayList<ArtifactItem>();
                }
                artifacts.add(art);
                m_definedPropertiesSources.put(prop, artifacts);
            }
        }
    }

    /**
     * Helper function for returned the node that is a child of the supplied node
     * and has the tag name provided.
     * @param n the node to search for children of
     * @param tag the tag to search for
     * @return the first child of n with the given tag name, or null if none exists
     */
    protected @Nullable Node lcl_getChildNodeWithTag(Node n, String tag) {
        NodeList nl = n.getChildNodes();
        for (int i = 0; i < nl.getLength(); ++i) {
            Node node = nl.item(i);
            if (node.getNodeName().equals(tag)) {
                return node;
            }
        }
        return null;
    }

}
