/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.paths;

import com.thetus.oss.maven.dependency.*;
import com.thetus.oss.maven.dependency.ILocalBinariesSource;
import com.thetus.oss.maven.hierarchy.IMavenProjectHierarchyBuilder;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import com.thetus.oss.maven.dependency.IDependencyResolver;
import com.thetus.oss.maven.dependency.MavenDependencyResolver;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.handler.ArtifactHandler;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuilder;
import org.apache.maven.project.ProjectBuildingException;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

/**
 * Basic implementation of IClasspathBuilder for use with maven projects. Capable of taking a project or
 * pom file and producing a classpath containing all available dependencies stored either in the local
 * repository or in a supplied project hierarchy. Filtering by scope and language is also possible, with
 * the default filtering looking for all java language dependencies with any scope.
 *
 * @author Samson "Zigurd" Bock - sbock@thetus.com
 */
public class MavenClasspathBuilder extends AMavenPathBuilder implements IClasspathBuilder {
    protected boolean m_useLocalBinaries = true;
    protected boolean m_useTestClasses = true;
    final protected Log m_log;

    protected List<ILocalBinariesSource> m_localBinariesSources = new ArrayList<ILocalBinariesSource>();

    /**
     * Initializes classpath builder with the maven supplied objects necessary for it to function, passes through
     * to {@link AMavenPathBuilder#AMavenPathBuilder(org.apache.maven.project.MavenProject, com.thetus.oss.maven.hierarchy.IMavenProjectHierarchyBuilder)}
     *
     * @param proj    the current project, used as the basis for determining what the classpath should be
     * @param builder object capable of building a maven project hierarchy
     * @param log     maven logger
     */
    public MavenClasspathBuilder(@NotNull final MavenProject proj, @NotNull final IMavenProjectHierarchyBuilder builder, @NotNull final Log log) {
        super(proj, builder);
        m_log = log;
    }

    /**
     * Initializes classpath builder with the maven supplied objects necessary for it to function, passes through
     * to {@link AMavenPathBuilder#AMavenPathBuilder(java.io.File, com.thetus.oss.maven.hierarchy.IMavenProjectHierarchyBuilder)}
     *
     * @param m_pomFile the pom file to build the basis project from
     * @param builder   object capable of building a maven project hierarchy
     * @param log       maven logger
     * @throws ProjectBuildingException if the pom file provided cannot be built into a valid project
     */
    public MavenClasspathBuilder(@NotNull final File m_pomFile, @NotNull final IMavenProjectHierarchyBuilder builder, @NotNull final Log log) throws ProjectBuildingException {
        super(m_pomFile, builder);
        m_log = log;
    }


    /**
     * {@inheritDoc}
     */
    @Override
    public void addLocalBinariesSource(ILocalBinariesSource h) {
        m_localBinariesSources.add(h);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setUseLocalBinaries(boolean tf) {
        m_useLocalBinaries = tf;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String buildClasspathString(String delim) {
        List<String> classpathStrings = getClasspathElements();
        StringBuilder builder = new StringBuilder();
        boolean first = true;
        for (String str : classpathStrings) {
            if (!first) {
                builder.append(delim);
            }
            first = false;
            builder.append(str);
        }
        return builder.toString();
    }

    /**
     * Builds up a list of all the elements in the classpath using the maven project this has as
     * the basis for what elements need to be included. If the builder has been configured to search out
     * the root aggregator, this will include the artifacts of the main project as well as all others in
     * the largest reactor it can find, as well as all dependencies (direct and transitive) of those artifacts.
     * If the builder has been set to use local binaries, it will replace .jar files with target/classes
     * directories where possible within the reactor.
     *
     * @return a list containing all the elements of a classpath, represented as strings
     */
    @Override
    public List<String> getClasspathElements() {
        if (m_project == null) {
            throw new IllegalStateException("It is baffling how a MavenClasspathBuilder was created with a null MavenProject. Failing out because we really need a non-null project");
        }

        IDependencyResolver resolverI = new MavenDependencyResolver(m_builder.getProjectBuilder(), m_builder.getSession() );
        final Set<Artifact> dependencies;
        List<String> dependencyStrings = new ArrayList<String>();
        MavenProjectHierarchy hierarchy;
        if (m_findRootAggregator) {
            hierarchy = m_builder.buildHierarchy(m_project, true, true);

            dependencies = resolverI.resolveDependencyArtifacts(hierarchy);
        } else {
            hierarchy = m_builder.buildHierarchy(m_project, false, false);
            MavenProject project = lcl_getBaseProject();
            dependencies = resolverI.resolveDependencyArtifacts(project);
        }

        for (MavenProject project : hierarchy.getProjectList()) {
            dependencies.add(project.getArtifact());
        }

        if (m_useLocalBinaries) {
            List<Artifact> foundDeps = new ArrayList<Artifact>();
            for (Artifact d : dependencies) {
                List<String> localBinaries = new ArrayList<String>();
                List<String> tmp = hierarchy.getLocalBinaryDirectoriesForCoordinate(d.getGroupId(), d.getArtifactId(), d.getVersion(), m_useTestClasses);
                if (tmp != null) {
                    localBinaries.addAll(tmp);
                    foundDeps.add(d);
                }
                for (ILocalBinariesSource src : m_localBinariesSources) {
                    List<String> tmp2 = src.getLocalBinaryDirectoriesForCoordinate(d.getGroupId(), d.getArtifactId(), d.getVersion(), m_useTestClasses);
                    if (tmp2 != null) {
                        localBinaries.addAll(tmp2);
                        foundDeps.add(d);
                    }
                }
                dependencyStrings.addAll(localBinaries);
            }

            for (Artifact d : foundDeps) {
                dependencies.remove(d);
            }
        }

        for (Artifact art : dependencies) {
            ArtifactHandler handler = art.getArtifactHandler();
            if (handler.getLanguage().equals("java")) {
                dependencyStrings.add(art.getFile().getAbsolutePath());
            }
        }

        return dependencyStrings;
    }

    @Override
    public void useTestClasses(boolean b) {
        m_useTestClasses = b;
    }
}
