/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.paths;

import com.thetus.oss.maven.hierarchy.IMavenProjectHierarchyBuilder;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import com.thetus.oss.maven.hierarchy.MavenProjectHierarchy;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectBuilder;
import org.apache.maven.project.ProjectBuildingException;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

/**
 * Basic implementation of {@link ISourcepathBuilder} for use with maven projects. Given a {@link ProjectBuilder} and {@link MavenSession}
 * it should be able to build a list of all relevant source paths for any supplied {@link MavenProject}. Source paths of a project include
 * src/main/* and target/generated-sources/
 *
 * @author Samson "Zigurd" Bock - sbock@thetus.com
 */
public class MavenSourcepathBuilder extends AMavenPathBuilder implements ISourcepathBuilder {
    List<String> m_excludes = new ArrayList<String>();
    boolean m_useTestSources = false;
    final Log m_log;

    /**
     * Passthrough to {@link AMavenPathBuilder#AMavenPathBuilder(org.apache.maven.project.MavenProject, com.thetus.oss.maven.hierarchy.IMavenProjectHierarchyBuilder)}
     *
     * @param proj basis project for determining what source paths to gather
     * @param log  maven logger
     */
    public MavenSourcepathBuilder(MavenProject proj, IMavenProjectHierarchyBuilder builder, Log log) {
        super(proj, builder);
        m_log = log;
    }

    /**
     * Passthrough to {@link AMavenPathBuilder#AMavenPathBuilder(java.io.File, com.thetus.oss.maven.hierarchy.IMavenProjectHierarchyBuilder)}
     *
     * @param m_pomFile file for constructing basis project used to determine what source paths to gather
     * @param log       maven logger
     * @throws ProjectBuildingException if the provided pom file cannot be used to create a valid maven project
     */
    public MavenSourcepathBuilder(File m_pomFile, IMavenProjectHierarchyBuilder builder, Log log) throws ProjectBuildingException {
        super(m_pomFile, builder);
        m_log = log;
    }

    /**
     * {@inheritDoc}
     *
     * @param delim the delimiter string to put between individual entries
     * @return
     */
    @Override
    public String buildSourcepathString(String delim) {
        List<String> elements = getSourcepathElements();
        StringBuilder builder = new StringBuilder();
        boolean first = true;
        for (String str : elements) {
            if (!first) {
                builder.append(delim);
            }
            first = false;
            builder.append(str);
        }
        return builder.toString();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public List<String> getSourcepathElements() {
        final List<String> elements = new ArrayList<String>();
        MavenProject baseProject = lcl_getBaseProject();
        MavenProjectHierarchy hierarchy = m_findRootAggregator ?
                m_builder.buildHierarchy(baseProject, true, true) :
                m_builder.buildHierarchy(baseProject, false, false);

        List<MavenProject> projects = hierarchy.getProjectList();
        for (MavenProject project : projects) {
            elements.addAll(project.getCompileSourceRoots());
            elements.add(project.getBasedir() + File.separator + "target" + File.separator + "generated-sources");
            if (m_useTestSources) {
                elements.addAll(project.getTestCompileSourceRoots());
            }
        }
        return elements;
    }

    @Override
    public void addExcludes(List<String> excludes) {
        for (String str : excludes) {
            m_excludes.add(str);
        }
    }

    @Override
    public void useTestSources(boolean b) {
        m_useTestSources = b;
    }
}
