/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.properties;

import org.jetbrains.annotations.Nullable;

/**
 * <p>
 * A data storage class used by the {@link PropertySetter}. It contains a property name and optional sources in which
 * a <code>PropertySetter</code> will look for a value for that property. It is usually defined in a pom.xml with the
 * following format:
 * </p>
 * <pre>
 * {@code
 * <propertyConfigurations>
 *     <propertyConfiguration>
 *         <targetProperty>thePropertyYouWantToSet</targetProperty>
 *         <systemProperty>java.sysprop.name</systemProperty>
 *         <environmentVariable>ENV_VARIABLE_SOURCE</environmentVariable>
 *         <propertiesFile>filename.properties</propertiesFile>
 *         <sourceProperty>mavenSourceProperty</sourceProperty>
 *         <defaultValue>MyDefaultValue</defaultValue>
 *     </propertyConfiguration>
 *     <propertyConfiguration>
 *         ...
 *
 *     </propertyConfiguration>
 * </propertyConfigurations>
 * }
 * </pre>
 * <p>
 *  A <code>PropertyConfiguration</code> can also be created programmatically using the provided setters.
 * </p>
 * <p>
 * All values other than <code>targetProperty</code> are optional and those present can be used by a
 * <code>PropertySetter</code> as possible sources to fill in the target property. If there is no valid
 * source among those provided, the target property will not be set at all. The different sources are used by the
 * <code>PropertySetter</code> as follows:
 * </p>
 * <ol>
 * <li>systemProperty: if the given system property is defined, store its value in targetProperty</li>
 * <li>environmentVariable: if the given environment variable is defined, store its value in targetProperty</li>
 * <li>propertiesFile: look for a file of the given name in local-configuration locations, if one is found, and
 * it contains a property with the name given in targetProperty, loads it into targetProperty</li>
 * <li>sourceProperty: if the given property exists in maven, store its value in targetProperty</li>
 * <li>defaultValue: if no other method resulted in a valid value, store the given value in targetProperty</li>
 * </ol>
 */
public class PropertyConfiguration {
    private String targetProperty;
    private String environmentVariable;
    private String propertiesFile;
    private String defaultValue;
    private String systemProperty;
    private String sourceProperty;
    private String ifDefined;

    public void setConfigFile(String propertiesFile) {
        this.propertiesFile = propertiesFile;
    }

    public void setSourceProperty(String sourceProperty) {
        this.sourceProperty = sourceProperty;
    }

    public void setDefaultValue(String defaultValue) {
        this.defaultValue = defaultValue;
    }

    /**
     * @return the property key to store a resultant value at
     */
    public String getPropertyName() {
        return targetProperty;
    }

    /**
     * @return the environment variable key to examine as a source for a property value
     */
    public String getEnvironmentVariableName() {
        return environmentVariable;
    }

    /**
     * @return the properties file to examine as a source for a property value
     */
    public String getConfigFile() {
        return propertiesFile;
    }

    /**
     * @return a default value to assign a property if no other sources provide a valid value
     */
    public String getDefaultValue() {
        return defaultValue;
    }

    /**
     * @return the name of a system property to examine as a source for a property value
     */
    public String getSystemPropertyName() {
        return systemProperty;
    }

    /**
     * @return the name of a maven property to examine as a source for a property value
     */
    public String getMavenPropertyName() {
        return sourceProperty;
    }

    /**
     * @return the name of a maven property upon which to base conditional processing of this configuration.
     * If the property is null or it is defined the configuration will be used, otherwise it will not.
     */
    public @Nullable String getIfDefined () {
        return ifDefined;
    }

    public void setPropertyName(String propertyName) {
        targetProperty = propertyName;
    }

    public void setEnvironmentVariableName(String environmentVariableName_arg) {
        environmentVariable = environmentVariableName_arg;
    }

    public void setIfDefined(String ifDefined) {
        this.ifDefined = ifDefined;
    }

    public void setSystemProperty(String str) {
        systemProperty = str;
    }
}
