/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.properties;

import com.thetus.oss.build.core.util.LocalConfigurationManager;
import org.apache.maven.plugin.logging.Log;
import org.apache.maven.project.MavenProject;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;

/**
 * <p>
 * A class that will attempt to set the values of a list or array of
 * {@link PropertyConfiguration <code>PropertyConfigurations</code>} by looking in a
 * variety of sources. If they are defined in each <code>PropertyConfiguration</code> the <code>PropertySetter</code>
 * will try to find a property in a system property (-D), an environment variable, a .properties file, a Maven property
 * set in the pom, and a default value in that order of precedence. If none of these can be found the
 * Property will not be set.
 * </p>
 * <p>
 * This class is intended for use in mojos that need to gather properties from places other than the pom.
 * </p>
 * <p>
 * In order to use this class you will need to define <code>&lt;propertyConfigurations&gt;</code> within the
 * <code>&lt;configuration&gt;</code> section of the plugin that will be using the
 * <code>PropertySetter</code> in the pom.
 * </p>
 * <p>
 * After instantiating the <code>PropertySetter</code>, calling <code>getProperties</code> will return a
 * {@link java.util.Properties} object containing the properties defined in the <code>PropertyConfigurations</code>.
 * </p>
 * <p>
 * For example, if, in your pom.xml, you define a <code>PropertyConfiguration</code> as:
 * </p>
 * <pre>
 * {@code
 * <propertyConfigurations>
 *     <propertyConfiguration>
 *         <targetProperty>foo</targetProperty>
 *         <environmentVariable>FOO_ENV_VAR</environmentVariable>
 *         <sourceProperty>foo.maven.property</sourceProperty>
 *         <defaultValue>bar</defaultValue>
 *     </propertyConfiguration>
 * </propertyConfigurations>
 * }
 * </pre>
 * <p>
 * And you define a the appropriate MojoParameters and MojoComponents in your mojo:
 * </p>
 * <pre>
 * <code>&#064;MojoParameter(alias = "propertyConfigurations")</code>
 * <code>private PropertyConfiguration[] m_propertyConfigurations;</code>
 *
 * <code>&#064;MojoComponent</code>
 * <code>private Maven m_maven;</code>
 * </pre>
 * <p>
 * Then, in your mojo, you can get those <code>Properties</code> like so:
 * </p>
 * <pre>
 * {@code
 * PropertySetter propertySetter = new PropertySetter(m_propertyConfigurations, m_maven, getLog());
 *
 * Properties properties = propertySetter.getProperties();
 * }
 * </pre>
 * <p>
 * If you do not have <code>FOO_ENV_VAR</code> or <code>foo.maven.property</code> set, the value of
 * <code>properties.getProperty("foo")</code> will be "bar". If you set the <code>FOO_ENV_VAR</code> to "baz" before
 * running the program but do not set <code>foo.maven.property</code>, the value will be "baz". If you execute the mojo
 * with <code>-Dfoo.maven.property=qux</code>, the value will be "qux".
 * </p>
 */
public class PropertySetter {

    /**
     * Enums corresponding to each of the sources that we can use to look for values to place in our target property.
     */
    private enum PropertySources { PROPERTY_FILES , SYSTEM_PROPERTIES , ENVIRONMENT_VARIABLES , MAVEN_PROPERTIES };

    /**
     * Hardcoded list used to represent the order of precedence for sources we can use to look for values.
     */
    private final PropertySources[] m_propertySources = new PropertySources[]{
        PropertySetter.PropertySources.SYSTEM_PROPERTIES ,
        PropertySetter.PropertySources.ENVIRONMENT_VARIABLES,
        PropertySetter.PropertySources.PROPERTY_FILES,
        PropertySetter.PropertySources.MAVEN_PROPERTIES
    };

    /**
     * Logger to be used for displaying warning information. Intended to be supplied by a Maven mojo, but
     * can be null, in which case log messages will simply be directed to std.out or std.err
     */
    protected Log m_log;

    /**
     * Maven project to be used as a source for current property values and a destination for new ones, if the
     * setProperties() function is used. Should generally be acquired from Maven as a mojo parameter, then passed
     * to this class.
     */
    protected MavenProject m_mavenProject;

    protected IPropertyContainer m_container;

    /**
     * The property configuration to act on, determines the sources to examine for property values, and the property
     * keys to assign them to. Consists of a number of different possible sources (environment variables, system
     * properties, maven properties, property files, etc) which if not null will be examined to determine a value
     * to be stored in the property at a key determined by {@link PropertyConfiguration#targetProperty}.
     */
    protected List<PropertyConfiguration> m_propertyConfigurations;

    /**
     * A default property file to look at for all configurations, if those configurations have not already supplied
     * a file to look at.
     */
    protected String m_propertiesFile;

    /**
     * A directory that the localConfigurationManager looks in to find the property file.
     */
    protected String m_groupName = null;

    /**
     * The local configuration manager to use to look for properties files, will be initialized
     * from best guesses if not explicityl provided.
     */
    protected LocalConfigurationManager m_localConfigurationManager = null;

    /**
     * Passthrough to main constructor, supplies null for defaultFile and groupName arguments.
     *
     * @param propConfigs properties configuration information
     * @param mproject source/destination for property values
     * @param logger logger to print warnings to, may be null
     */
    @SuppressWarnings("unused")
    public PropertySetter( PropertyConfiguration[] propConfigs, MavenProject mproject , @Nullable Log logger ) {
        this(Arrays.asList(propConfigs),mproject,null, null,logger);
    }

    /**
     * Passthrough to main constructor, supplies null for defaultFile and groupName arguments.
     *
     * @param propConfigs properties configuration information
     * @param mproject source/destination for property values
     * @param logger logger to print warnings to, may be null
     */
    @SuppressWarnings("unused")
    public PropertySetter( List<PropertyConfiguration> propConfigs, MavenProject mproject , @Nullable Log logger ) {
        this(propConfigs, mproject, null, null, logger);
    }

     /**
     * Passthrough to main constructor, supplies null for the groupName argument.
     *
     * @param propConfigs properties configuration information
     * @param mproject source/destination for property values
     * @param defaultFile property file to use if none is specified by an individual configuration.
     * @param logger logger to print warnings to, may be null
     */
    @SuppressWarnings("unused")
    public PropertySetter( List<PropertyConfiguration> propConfigs, MavenProject mproject, String defaultFile , @Nullable Log logger ) {
        this(propConfigs,mproject,null, null,logger);
    }

    /**
     * Passthrough to main constructor, Takes in an array of PropertyConfigurations instead of a list.
     *
     * @param propConfigs properties configuration information
     * @param mproject source/destination for property values
     * @param defaultFile property file to use if none is specified by an individual configuration, may be null
     * @param groupName the group name that the local configuration manager looks in for the property file
     * @param logger logger to print warnings to, may be null
     */
    @SuppressWarnings("unused")
    public PropertySetter( PropertyConfiguration[] propConfigs, MavenProject mproject , @Nullable String defaultFile, @Nullable String groupName, @Nullable Log logger ) {
        this(Arrays.asList(propConfigs),mproject,defaultFile, groupName,logger);
    }

    public PropertySetter( List<PropertyConfiguration> propConfigs, MavenProject mproject , @Nullable String defaultFile, @Nullable String groupName, @Nullable Log logger ) {
        this(propConfigs, new PropsContainer(mproject.getProperties(), mproject.getArtifactId()), defaultFile, groupName, logger, mproject);
    }

    /**
     * Constructor simply initializes variables, does not have any significant logic itself. What is done with the supplied
     * parameters is determined by calls to {@link PropertySetter#getProperties()} or
     * {@link PropertySetter#setMavenProperties()}.
     *
     * @param propConfigs properties configuration information
     * @param container source/destination for property values
     * @param defaultFile property file to use if none is specified by an individual configuration, may be null
     * @param groupName the group name that the local configuration manager looks in for the property file
     * @param logger logger to print warnings to, may be null
     * @param project maven project
     */
    public PropertySetter(List<PropertyConfiguration> propConfigs, IPropertyContainer container,
                          @Nullable String defaultFile, @Nullable String groupName, @Nullable Log logger,
                          @Nullable MavenProject project) {
        m_propertyConfigurations = propConfigs;
        m_container = container;
        m_mavenProject = project;
        m_log = logger;
        m_propertiesFile = defaultFile;
        m_groupName = groupName;
    }

    /**
     * For each entry in {@link PropertySetter#m_propertyConfigurations} examines each property source (System Properties,
     * Environment Variables, Property Files, and Maven Properties) in order, as determined by the {@link PropertySetter#m_propertySources}
     * array. When a source is found that is non-null and contains a valid value, it is taken, and stored in the property
     * at the key determined by {@link PropertyConfiguration#targetProperty}. If none is found and the {@link PropertyConfiguration#defaultValue}
     * is set, that is used instead. If none of those apply, the property is not changed.
     *
     * @return a Properties object containing all the property key-value pairs created from the supplied configuration values.
     */
    public Properties getProperties () {
        Properties p = new Properties();
        lcl_processProperties(p);
        return p;
    }

    public void setLocalConfigurationManager(LocalConfigurationManager localConfigurationManager) {
        m_localConfigurationManager = localConfigurationManager;
    }

    /**
     * As {@link PropertySetter#getProperties()}, but instead of returning a {@link Properties} object,
     * this function instead sets values in the current Maven Properties. These properties should remain set throughout the lifecycle
     * of the project.
     */
    public void setMavenProperties () {
        lcl_processProperties(null);
    }

    /**
     * Helper function to perform the logic for {@link PropertySetter#getProperties} and {@link PropertySetter#setMavenProperties()},
     * see the former for description of the logic used.
     * @param p the object to store property values in, if any. If not supplied the {@link PropertySetter#m_container} member variable is used instead.
     */
    protected void lcl_processProperties (@Nullable Properties p) {
        for ( PropertyConfiguration config : m_propertyConfigurations) {
            String result = null;
            if ( config.getPropertyName() != null ) {
                if ( lcl_checkIfDefined(config,p) ) {
                    for ( PropertySources source : m_propertySources) {
                        switch(source) {
                            case SYSTEM_PROPERTIES:
                                result = lcl_getPropertyFromSystemProperties(config);
                                break;
                            case ENVIRONMENT_VARIABLES:
                                result = lcl_getPropertyFromEnvironmentVariables(config);
                                break;
                            case PROPERTY_FILES:
                                String fileName = config.getConfigFile() != null ? config.getConfigFile() : m_propertiesFile;
                                result = lcl_getPropertyFromPropertyFiles(config, fileName);
                                break;
                            case MAVEN_PROPERTIES:
                                result = lcl_getPropertyFromMavenProperties(config);
                                break;
                            default:
                                if ( m_log != null ) { m_log.warn("unknown source type"); }
                        }
                        if ( result != null ) {
                            break;
                        }
                    }
                    if ( result == null ) {
                        String tmpResult = config.getDefaultValue();
                        if ( tmpResult != null ) {
                            int currentIndex = 0;

                            StringBuilder builder = new StringBuilder();
                            while ( tmpResult.indexOf("${",currentIndex) >= 0 ) {
                                int n = tmpResult.indexOf("${",currentIndex);
                                if ( n - currentIndex > 0 ) { builder.append(tmpResult.substring(currentIndex,n)); }

                                int end = tmpResult.indexOf("}",n);
                                String var = tmpResult.substring(n+2,end);
                                currentIndex = end+1;

                                String val;// = null;
                                if ( p != null ) {
                                    val = p.getProperty(var);
                                } else {
                                    val = m_container.getProperties().getProperty(var);
                                }
                                if ( val == null ) {
                                    val = System.getProperty(var);
                                }
                                if ( val == null && var.toLowerCase().startsWith("env.")) {
                                    val = System.getenv(var.substring(4));
                                }

                                if ( val != null ) {
                                    builder.append(val);
                                }
                            }

                            if ( tmpResult.length() > currentIndex ) { builder.append(tmpResult.substring(currentIndex)); }

                            result = builder.toString();
                            System.out.println("RESULT: " + result);
                        }
                    }

                    if ( result != null ) {
                        if ( m_log != null ) {
                            m_log.info("Setting property \"" + config.getPropertyName() + "\" to \"" + result + "\"");
                        }
                        if ( p != null ) {
                            p.setProperty(config.getPropertyName(),result);
                        } else {
                            m_container.getProperties().setProperty(config.getPropertyName(),result);
                        }
                    }
                }
                else{
                    m_log.info("NOT DEFINED: " + config.getIfDefined());
                }
            }
        }
    }

    /**
     * Helper function for checking the ifDefined element of the config. Looks to find if the ifDefined property
     * exists in the maven project's properties or the global system properties.
     * @param config the configuration to check
     * @param props the properties object that we are using to store our results
     * @return true if the ifDefined property is defined (or if none is supplied), false otherwise
     */
    private boolean lcl_checkIfDefined(PropertyConfiguration config,@Nullable Properties props) {
        if ( config.getIfDefined() != null ) {
            return m_container.getProperties().getProperty(config.getIfDefined()) != null ||
                    System.getProperty(config.getIfDefined()) != null ||
                    (props != null && props.getProperty(config.getIfDefined()) != null);
        }
        return true;
    }

    /**
     * Internal method for acquiring a property from a System property, if one of the appropriate name exists. If a
     * system property to check is not explicitly provided, this function will look for one with the same name as
     * the target property.
     * @param config (not null) the {@link PropertyConfiguration} that contains the parameters acquired from the pom.xml file
     * @return A string containing the value of the System property of the appropriate name, or <code>null</code> if
     * no appropriate System property exists.
     */
    private @Nullable
    String lcl_getPropertyFromSystemProperties( @NotNull PropertyConfiguration config ) {
        String sysPropName = config.getSystemPropertyName();
        String sysProp = System.getProperty(sysPropName != null ? sysPropName : config.getPropertyName());
        return sysProp;
    }
    /**
     * Internal method for acquiring a property from an environment variable, if one of the appropriate name exists. If
     * an environment variable is not explicitly provided, this function will look for one with the same name as the
     * target property (unlikely though it is that an environment variable so named will exist)
     * @param config (not null) the {@link PropertyConfiguration} that contains the parameters acquired from the pom.xml file
     * @return A string containing the value of the environment variable of the appropriate name, or <code>null</code> if
     * no appropriate environment variable exists.
     */
    private @Nullable String lcl_getPropertyFromEnvironmentVariables ( @NotNull PropertyConfiguration config ) {
        String envVarName = config.getEnvironmentVariableName();
        String envVar = System.getenv(envVarName != null ? envVarName : config.getPropertyName());
        return envVar;
    }

    /**
     * Internal method for acquiring a property from property files. This method depends on the
     * {@link LocalConfigurationManager} for acquiring the appropriate files. Once those files have been acquired,
     * properties are extracted from them in order of precedence, and the appropriate property value is returned.
     * If the user has not supplied a config file name, this method will simply return <code>null</code>.
     * @param config (not null) the {@link PropertyConfiguration} that contains the parameters acquired from the pom.xml file
     * @param fileName the name of the properties file (not its full path), if null, the function immediately returns
     * @return A string containing the value of the property found in one of the property files, or <code>null</code>
     * if no appropriate property was found, or if no appropriate files were found.
     */
    private @Nullable String lcl_getPropertyFromPropertyFiles ( @NotNull PropertyConfiguration config, String fileName ) {
        if ( fileName == null ) { return null; }
        LocalConfigurationManager manager = m_localConfigurationManager;
        if ( manager == null ) {
            if (m_groupName != null){
                if ( m_mavenProject != null ) {
                    manager = new LocalConfigurationManager(m_mavenProject.getBasedir(),m_groupName);
                } else {
                    manager = new LocalConfigurationManager(m_groupName);
                }
            }else {
                if ( m_mavenProject != null ) {
                    manager = new LocalConfigurationManager(m_mavenProject.getBasedir(),m_container.GetGroupName());
                } else {
                    manager = new LocalConfigurationManager(m_container.GetGroupName());
                }
            }
        }
        List<File> configFiles = manager.findAllLocalConfigFile(fileName);
        Properties properties = new Properties();
        for ( File f : configFiles ) {
            try {
                properties.load(new BufferedInputStream(new FileInputStream(f)));
            } catch (IOException e) {
                if ( m_log != null ) {
                    m_log.warn("File returned from localConfigurationManager did not exist");
                    m_log.warn("File path: \"" + f.getAbsolutePath() + "\"");
                }
            }
        }
        String configFileValue = properties.getProperty(config.getPropertyName());
        return configFileValue;
    }

    /**
     * Internal method for acquiring a property from a Maven property, if one of the appropriate name exists
     * @param config (not null) the {@link PropertyConfiguration} that contains the parameters acquired from the pom.xml file
     * @return A string containing the value of the Maven property of the appropriate name, or <code>null</code> if
     * no appropriate Maven property exists.
     */
    private String lcl_getPropertyFromMavenProperties ( @NotNull PropertyConfiguration config ) {
        Properties mavenProperties = m_container.getProperties();
        String mavenSourceName = config.getMavenPropertyName();
        if ( mavenProperties != null && mavenSourceName != null ) {
            return mavenProperties.getProperty(mavenSourceName);
        }
        return null;
    }

}
