/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.repository;

import com.thetus.oss.maven.execution.MavenExecutor;
import com.thetus.oss.maven.execution.MavenExecutor;
import org.apache.maven.Maven;
import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.DefaultArtifact;
import org.apache.maven.artifact.handler.DefaultArtifactHandler;
import org.apache.maven.artifact.metadata.ArtifactMetadataRetrievalException;
import org.apache.maven.artifact.metadata.ArtifactMetadataSource;
import org.apache.maven.artifact.versioning.ArtifactVersion;
import org.apache.maven.execution.MavenExecutionResult;
import org.apache.maven.execution.MavenSession;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.List;
import java.util.Properties;

/**
 * This class provides the ability to communicate with remote maven repositories
 * in order to perform a few non-standard tasks, namely getting a list of
 * versions available for a certain artifact and downloading specific artifacts
 * directly to the local repository.
 *
 * Internally, this helper class makes use of a piece of maven voodoo called an
 * <code>ArtifactMetadataSource</code>. This is all well and good, but it should
 * be noted that it is technically deprecated, and this class will need updating
 * come maven 4, if we end up migrating there when it eventually comes out.
 *
 * @author Zig - sbock@thetus.com
 */
public class RepositoryCommunicator {
    MavenSession m_session;
    Maven m_maven;
    ArtifactMetadataSource m_metadataSource;


    public RepositoryCommunicator(@NotNull MavenSession session, @NotNull Maven maven, @NotNull ArtifactMetadataSource metadataSource) {
        m_session = session;
        m_maven = maven;
        m_metadataSource = metadataSource;
    }

    /**
     * Acquires a list of available versions of an artifact with the supplied groupId and artifactId. If the artifact
     * in question is not a jar, use the {@link #getAvailableArtifactVersions(String, String, String)} function and
     * specify the packaging type instead.
     *
     * @param groupId groupId of the artifact to retrieve version information about
     * @param artifactId artifactId of the artifact to retrieve version information about
     * @return a list of versions available from the remote repos
     */
    @Nullable
    public List<ArtifactVersion> getAvailableArtifactVersions (@NotNull String groupId,@NotNull String artifactId) {
        return getAvailableArtifactVersions(groupId,artifactId,"jar");
    }

    /**
     * Acquires a list of available versions of an artifact with the supplied groupId, artifactId and packaging type.
     *
     * @param groupId groupId of the artifact to retrieve version information about
     * @param artifactId artifactId of the artifact to retrieve version information about
     * @param type the pacakaging type of the artifact to retrieve version information about
     * @return a list of versions available from the remote repos
     */
    @Nullable
    public List<ArtifactVersion> getAvailableArtifactVersions (@NotNull String groupId,@NotNull String artifactId,@NotNull String type) {
        Artifact tmpArtifact = new DefaultArtifact(groupId,artifactId,"[0.0,)","compile",type,null,new DefaultArtifactHandler());
        try {
            return m_metadataSource.retrieveAvailableVersions(tmpArtifact,m_session.getLocalRepository(),m_session.getRequest().getRemoteRepositories());
        } catch (ArtifactMetadataRetrievalException e) {
            return null;
        }
    }

    /**
     * Downloads an artifact specified by the give groupId, artifactId, version and packaging type into the local
     * repository. To access the actual files downloaded, you will need to communicate with the local repository.
     *<br />
     * Internally, this calls into the maven-dependency-plugin rather than reimplement the functionality in that
     * plugin. As such, this makes use of our magical {@link com.thetus.oss.maven.execution.MavenExecutor}.
     *<br />
     * This function is fairly fragile, and should be used with care, if you have a different way to accomplish the
     * same thing, you might consider doing that instead. As a short term and/or hacky solution, however, this
     * should serve.
     *
     * @param groupId groupId of the artifact to download
     * @param artifactId artifactId of the artifact to download
     * @param version version of the artifact to download
     * @param packaging the packaging type of the artifact to download
     * @return true if the artifact was successfully downloaded, false otherwise
     */
    public boolean getArtifact (@NotNull String groupId,@NotNull String artifactId,@NotNull String version,@NotNull String packaging){
        MavenExecutor executor = new MavenExecutor(m_session,m_maven);
        executor.addGoal("org.apache.maven.plugins:maven-dependency-plugin:2.2:get");
        Properties userProperties = m_session.getRequest().getUserProperties();
        userProperties.setProperty("repoUrl",m_session.getRequest().getRemoteRepositories().get(0).getUrl());
        userProperties.setProperty("artifact",groupId + ":" + artifactId + ":" + version);
        userProperties.setProperty("packaging",packaging);

        m_session.getRequest().setUserProperties(userProperties);
        MavenExecutionResult result = executor.execute();
        if ( result.hasExceptions() ){
            return false;
        }
        return true;
    }
}
