/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.util;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.DefaultArtifact;
import org.apache.maven.artifact.handler.DefaultArtifactHandler;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Dependency;
import org.apache.maven.project.*;

/**
 * Helper class that provides utility functions for manipulating maven {@link Dependency}, {@link MavenProject} , and {@link Artifact} objects.
 *
 * @author Samson "Zigurd" Bock - sbock@thetus.com
 */
public class DependencyHelper {
    /**
     * Converts a {@link Dependency} object into an {@link Artifact} object using the Dependency's groupId,artifactId,version,scope,type and classifier,
     * and a {@link DefaultArtifactHandler}
     * @param d the dependency to be converted, not modified by function
     * @return a new {@link DefaultArtifact} wrapping the dependency parameter's information
     */
    public static Artifact convertDependencyToArtifact (Dependency d){
        return new DefaultArtifact(d.getGroupId(),d.getArtifactId(),d.getVersion(),d.getScope(),d.getType(),d.getClassifier(),new DefaultArtifactHandler());
    }

    /**
     * Creates a project representing the provided dependency. To do so, this function looks for an {@link Artifact} matching the {@link Dependency}
     * in the local repository, and uses a {@link ProjectBuilder} to construct a {@link MavenProject} from that.
     * @param d the dependency to build a project from
     * @param pb the project builder to do the actual constructing, will likely originate from a maven component
     * @param ms the maven session used to acquire the local {@link org.apache.maven.artifact.repository.ArtifactRepository} and project building information
     * @return a new {@link MavenProject} based on the local repository artifact corresponding to the provided {@link Dependency}
     * @throws ProjectBuildingException if the project could not be built
     */
    public static MavenProject createProjectFromDependency (Dependency d,ProjectBuilder pb,MavenSession ms) throws ProjectBuildingException {
        Artifact art = ms.getLocalRepository().find( convertDependencyToArtifact(d) );
        return pb.build(art,ms.getProjectBuildingRequest()).getProject();
    }

    /**
     * Converts a {@link Artifact} object into a {@link Dependency} object using the Artifact's groupId,artifactId,version,classifier,scope
     * @param art the artifact to convert
     * @return the new {@link Dependency} object
     */
    public static Dependency convertArtifactToDependency (Artifact art){
        Dependency d = new Dependency();
        d.setArtifactId(art.getArtifactId());
        d.setGroupId(art.getGroupId());
        d.setVersion(art.getVersion());
        d.setClassifier(art.getClassifier());
        d.setScope(art.getScope());
        return d;
    }
}
