/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.util;

import org.apache.maven.model.Plugin;
import org.apache.maven.model.PluginExecution;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.util.xml.Xpp3Dom;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import sun.reflect.generics.reflectiveObjects.NotImplementedException;

import java.util.ArrayList;
import java.util.List;

/**
 * A collection of small helper functions for commonly needed operations on {@link MavenProject}'s and collections
 * of the same. Functions exist for filtering based on packaging type, language, use of plugins, etc.
 *
 * @author Samson "Zigurd" Bock - sbock@thetus.com
 */
public class MavenProjectUtils {

    /**
     * Helper function for testing if a given project uses a specified plugin goal
     * @param p project to check
     * @param pluginArtifactId plugin artifact to look for
     * @param goalName goal of plugin to look for
     * @return true if the project uses the specified plugin goal, false otherwise
     */
    public static boolean doesProjectUsePluginGoal(MavenProject p,@NotNull String pluginArtifactId,@Nullable String goalName){
        List<Plugin> plugins = p.getBuildPlugins();
        for ( Plugin plugin : plugins ) {
            if ( pluginArtifactId.equals(plugin.getArtifactId()) ) {
                List<PluginExecution> executions = plugin.getExecutions();
                if ( executions != null ) {
                    for ( PluginExecution execution : executions ) {
                        if ( goalName != null ){
                            List<String> goals = execution.getGoals();
                            if ( goals != null ) {
                                for ( String goal : goals ) {
                                    if ( goalName.equals(goal) ) {
                                        return true;
                                    }
                                }
                            }
                        }
                        else{
                            return true;
                        }
                    }
                }
            }
        }
        return false;
    }

    /**
     * Gets a plugin's configuration as a list of Xpp3Dom nodes (one per execution profile)
     * @param p the project to get configurations from
     * @param pluginArtifactID the plugin artifact ID to get configuration data for
     * @param goalName the goal of the plugin to get configuration data for
     * @return a non-null list of base nodes for the configuration sections of the plugin
     */
    public static @NotNull List<Xpp3Dom> getPluginConfigurations(@NotNull MavenProject p,@NotNull String pluginArtifactID ,@NotNull String goalName) {
        List<Xpp3Dom> ret = new ArrayList<Xpp3Dom>();

        for ( Plugin plugin : p.getBuildPlugins() ) {
            if ( pluginArtifactID.equals(plugin.getArtifactId()) ) {
                List<PluginExecution> executions = plugin.getExecutions();
                if ( executions != null ) {
                    for ( PluginExecution execution : executions ) {
                        List<String> goals = execution.getGoals();
                        if ( goals != null ) {
                            for ( String goal : goals ) {
                                if ( goalName.equalsIgnoreCase(goal) ) {
                                    ret.add(p.getGoalConfiguration(plugin.getGroupId(),plugin.getArtifactId(),execution.getId(),goal));
                                }
                            }
                        }
                    }
                }
            }
        }
        return ret;
    }

    public static @Nullable String getPluginConfigurationValueByPath(
            @NotNull MavenProject p,@NotNull String pluginArtifactID, @NotNull String goalName,@NotNull String path)
    {
        List<Xpp3Dom> configurationItems = getPluginConfigurations(p,pluginArtifactID,goalName);
        for ( Xpp3Dom node : configurationItems ){
            //node.
            throw new NotImplementedException();
        }
        return null;
    }

    public static @NotNull List<MavenProject> getProjectsWithLanguages (@NotNull List<MavenProject> projects,@NotNull final String[] languages){
        return filterProjects(projects,new ProjectFilter() {
            @Override
            boolean isAccepted(MavenProject p) {
                if ( p.getArtifact() != null && p.getArtifact().getArtifactHandler() != null ){
                    for ( String language : languages ){
                        if ( p.getArtifact().getArtifactHandler().getLanguage().equalsIgnoreCase(language) ){
                            return true;
                        }
                    }
                }
                else{
                    System.out.println("Project without artifact or artifactHandler encountered while attempting to filter by language");
                }
                return false;
            }
        });
    }

    public static @NotNull List<MavenProject> getProjectsWithPackagingType (@NotNull List<MavenProject> projects,@NotNull final String packagingType){
        return filterProjects(projects,new ProjectFilter() {
            @Override
            boolean isAccepted(MavenProject p) {
                return p.getPackaging().equalsIgnoreCase(packagingType);
            }
        });
    }

    public static @NotNull List<MavenProject> getProjectsWithPackagingTypes (@NotNull List<MavenProject> projects,@NotNull final String[] packagingType){
        return filterProjects(projects,new ProjectFilter() {
            @Override
            boolean isAccepted(MavenProject p) {
                for ( String ptype : packagingType ){
                    if ( p.getPackaging().equalsIgnoreCase(ptype) ){
                        return true;
                    }
                }
                return false;
            }
        });
    }

    public static @NotNull List<MavenProject> filterProjects (@NotNull List<MavenProject> projects,@NotNull ProjectFilter filter){
        List<MavenProject> ret = new ArrayList<MavenProject>();
        for ( MavenProject p : projects ){
            if ( filter.isAccepted(p) ){
                ret.add(p);
            }
        }
        return ret;
    }

    public static abstract class ProjectFilter{
        abstract boolean isAccepted(MavenProject p);
    }
}
