/*
   Copyright(c) 2002-2011 Thetus Corporation. www.thetus.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.thetus.oss.maven.util;

import org.apache.commons.io.FilenameUtils;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.util.regex.Pattern;

/**
 * Created by IntelliJ IDEA.
 * User: sbock
 * Date: 5/27/11
 * Time: 3:46 PM
 */
public class RelativeFilenameUtils {
    /**
     * Based on the submission of user Don of StackOverflow
     * http://stackoverflow.com/questions/204784/how-to-construct-a-relative-path-in-java-from-two-absolute-paths-or-urls
     *
     * Get the relative path from one file to another
     * If one of the provided resources does not exist, it is assumed to be a file unless it ends with '/' or
     * '\'.
     *
     * @param targetPath the path to transform into a relative path
     * @param basePath the path used as a basis for the transformation
     * @return the relative path in string form, or null if no relative path exists
     */
    public static @Nullable
    String getRelativePath(String targetPath, String basePath) {
        if ( targetPath.equals(basePath) ){ return "."; }
        // Normalize the paths
        String normalizedTargetPath = FilenameUtils.normalizeNoEndSeparator(targetPath);
        String normalizedBasePath = FilenameUtils.normalizeNoEndSeparator(basePath);

        // Restore the path separators to their previous state
        if (File.separator.equals("/")) {
            normalizedTargetPath = FilenameUtils.separatorsToUnix(normalizedTargetPath);
            normalizedBasePath = FilenameUtils.separatorsToUnix(normalizedBasePath);

        } else if (File.separator.equals("\\")) {
            normalizedTargetPath = FilenameUtils.separatorsToWindows(normalizedTargetPath);
            normalizedBasePath = FilenameUtils.separatorsToWindows(normalizedBasePath);

        }

        String[] basePathElements = normalizedBasePath.split(Pattern.quote(File.separator));
        String[] targetPathElements = normalizedTargetPath.split(Pattern.quote(File.separator));

        // First get all the common elements. Store them as a string,
        // and count how many of them there are.
        StringBuffer sharedPath = new StringBuffer();

        int sharedIndex = 0;
        while (sharedIndex < targetPathElements.length && sharedIndex < basePathElements.length
                && targetPathElements[sharedIndex].equals(basePathElements[sharedIndex])) {
            sharedPath.append(targetPathElements[sharedIndex] + File.separator);
            sharedIndex++;
        }

        if (sharedIndex == 0) {
            return null;
        }

        // The number of directories we have to backtrack depends on whether the base is a file or a dir
        // For example, the relative path from
        //
        // /foo/bar/baz/gg/ff to /foo/bar/baz
        //
        // ".." if ff is a file
        // "../.." if ff is a directory
        //
        // The following is a heuristic to figure out if the base refers to a file or dir. It's not perfect, because
        // the resource referred to by this path may not actually exist, but it's the best we can do
        boolean baseIsFile = true;

        File baseResource = new File(normalizedBasePath);

        if (baseResource.exists()) {
            baseIsFile = baseResource.isFile();

        } else if (basePath.endsWith(File.separator)) {
            baseIsFile = false;
        }

        StringBuffer relative = new StringBuffer();

        if (basePathElements.length != sharedIndex) {
            int numDirsUp = baseIsFile ? basePathElements.length - sharedIndex - 1 : basePathElements.length - sharedIndex;

            for (int i = 0; i < numDirsUp; i++) {
                relative.append(".." + File.separator);
            }
        }


        relative.append(normalizedTargetPath.substring( Math.min(sharedPath.length(), normalizedTargetPath.length() - 1)));
        return relative.toString();
    }
}
